from __future__ import annotations


class ReflectionBlockIntegratorStills:
    """A class to perform the integration."""

    def __init__(self, params, experiments, reference, extractor=None):
        """Initialise the integrator."""
        from dials.algorithms import shoebox

        # Ensure we have 1 experiment at the moment
        assert len(experiments) == 1
        assert extractor is not None

        # Save the parameters
        self.params = params
        self.experiments = experiments
        self.extractor = extractor

        # Create the shoebox masker
        n_sigma = params.integration.shoebox.n_sigma
        assert n_sigma > 0
        self._mask_profiles = shoebox.MaskerEmpirical(
            experiments[0], reference=reference
        )

    def integrate(self):
        """Integrate all the reflections."""
        from dials.algorithms.shoebox import MaskCode
        from dials.array_family import flex

        result = flex.reflection_table()
        for indices, reflections in self.extractor:
            self._mask_profiles(reflections, None)
            reflections.integrate(self.experiments[0])
            bg_code = MaskCode.Valid | MaskCode.BackgroundUsed
            fg_code = MaskCode.Valid | MaskCode.Foreground
            n_bg = reflections["shoebox"].count_mask_values(bg_code)
            n_fg = reflections["shoebox"].count_mask_values(fg_code)
            reflections["n_background"] = n_bg
            reflections["n_foreground"] = n_fg
            del reflections["shoebox"]
            del reflections["rs_shoebox"]
            result.extend(reflections)
        assert len(result) > 0
        result.sort("miller_index")
        return result


class IntegratorStills:
    """Integrate reflections"""

    def __init__(self, params, exlist, reference=None, predicted=None, shoeboxes=None):
        """Initialise the script."""

        assert reference is not None

        # Load the extractor based on the input
        if shoeboxes is not None:
            extractor = self._load_extractor(shoeboxes, params, exlist)
        else:
            if predicted is None:
                predicted = self._predict_reflections(params, exlist)
                # predicted = self._filter_reflections(params, exlist, predicted) # FIXME

            predicted = self._match_with_reference(predicted, reference)

            import math

            from annlib_ext import AnnAdaptor

            from dials.array_family import flex

            matcheddata = predicted.select(
                predicted.get_flags(predicted.flags.reference_spot)
            )

            A = AnnAdaptor(matcheddata["xyzcal.mm"].as_double(), 3, 10)
            A.query(predicted["xyzcal.mm"].as_double())

            bboxes = flex.int6()
            for i, ref in enumerate(predicted):
                nn_pred = [matcheddata[A.nn[i * 10 + j]] for j in range(10)]
                nn_ref = [
                    reference[reference["miller_index"].first_index(r["miller_index"])]
                    for r in nn_pred
                ]

                max_x = max([r["bbox"][1] - r["bbox"][0] for r in nn_ref])
                max_y = max([r["bbox"][3] - r["bbox"][2] for r in nn_ref])

                panel = exlist[ref["id"]].detector[ref["panel"]]
                imgsize_x, imgsize_y = panel.get_image_size()

                x1 = int(math.floor(ref["xyzcal.px"][0] - (max_x / 2)))
                x2 = int(math.ceil(ref["xyzcal.px"][0] + (max_x / 2)))
                y1 = int(math.floor(ref["xyzcal.px"][1] - (max_y / 2)))
                y2 = int(math.ceil(ref["xyzcal.px"][1] + (max_y / 2)))

                if x1 < 0:
                    x1 = 0
                if y1 < 0:
                    y1 = 0

                if x2 > imgsize_x:
                    x2 = imgsize_x
                if y2 > imgsize_y:
                    y2 = imgsize_y

                bboxes.append((x1, x2, y1, y2, 0, 1))

            predicted["bbox"] = bboxes

            extractor = self._create_extractor(params, exlist, predicted)

        # Initialise the integrator
        self._integrator = ReflectionBlockIntegratorStills(
            params, exlist, reference, extractor
        )

    def integrate(self):
        """Integrate the reflections."""
        return self._integrator.integrate()

    def _match_with_reference(self, predicted, reference):
        """Match predictions with reference spots."""

        from dials.algorithms.spot_finding.spot_matcher import SpotMatcher
        from dials.util.command_line import Command

        Command.start("Matching reference spots with predicted reflections")
        match = SpotMatcher(max_separation=1)
        rind, pind = match(reference, predicted)
        h1 = predicted.select(pind)["miller_index"]
        h2 = reference.select(rind)["miller_index"]
        mask = h1 == h2
        predicted.set_flags(pind.select(mask), predicted.flags.reference_spot)
        Command.end(
            "Matched %d reference spots with predicted reflections" % mask.count(True)
        )
        return predicted

    def _load_extractor(self, filename, params, exlist):
        """Load the shoebox extractor."""
        from dials.model.serialize.reflection_block import ReflectionBlockExtractor

        assert len(exlist) == 1
        imageset = exlist[0].imageset
        return ReflectionBlockExtractor(
            filename, params.integration.shoebox.block_size, imageset
        )

    def _create_extractor(self, params, exlist, predicted):
        """Create the extractor."""
        from dials.model.serialize.reflection_block import ReflectionBlockExtractor

        assert len(exlist) == 1
        imageset = exlist[0].imageset
        return ReflectionBlockExtractor(
            "shoebox.dat", params.integration.shoebox.block_size, imageset, predicted
        )

    def _predict_reflections(self, params, experiments):
        """Predict all the reflections."""
        from dials.array_family import flex

        result = flex.reflection_table()
        for i, experiment in enumerate(experiments):
            predicted = flex.reflection_table.from_predictions(experiment)
            predicted["id"] = flex.int(len(predicted), i)
            result.extend(predicted)
        return result

    def _filter_reflections(self, params, experiments, reflections):
        """Filter the reflections to integrate."""
        from dials.algorithms import filtering
        from dials.array_family import flex
        from dials.util.command_line import Command

        # Set all reflections which overlap bad pixels to zero
        Command.start("Filtering reflections by detector mask")
        if experiments[0].scan is None:
            array_range = 1
        else:
            array_range = experiments[0].scan.get_array_range()
        mask = filtering.by_detector_mask(
            reflections["bbox"],
            experiments[0].imageset.get_raw_data(0)[0] >= 0,
            array_range,
        )
        reflections.del_selected(not mask)
        Command.end(f"Filtered {len(reflections)} reflections by detector mask")

        # Filter the reflections by zeta
        min_zeta = params.integration.filter.by_zeta
        if min_zeta > 0:
            Command.start(f"Filtering reflections by zeta >= {min_zeta:f}")
            zeta = reflections.compute_zeta(experiments[0])
            reflections.del_selected(flex.abs(zeta) < min_zeta)
            n = len(reflections)
            Command.end("Filtered %d reflections by zeta >= %f" % (n, min_zeta))
            return reflections



from __future__ import annotations

import concurrent.futures
import logging
import math

import libtbx
from libtbx.phil import parse

from dials.array_family import flex
from dials.util import tabulate
from dials.util.mp import available_cores

logger = logging.getLogger(__name__)


class CentroidOutlier:
    """Base class for centroid outlier detection algorithms"""

    def __init__(
        self,
        cols=None,
        min_num_obs=20,
        separate_experiments=True,
        separate_panels=True,
        separate_images=False,
        block_width=None,
        nproc=1,
    ):

        # column names of the data in which to look for outliers
        if cols is None:
            cols = ["x_resid", "y_resid", "phi_resid"]
        self._cols = cols

        # minimum number of observations per panel below which all reflections will
        # be marked as potential outliers
        self._min_num_obs = min_num_obs

        # whether to do outlier rejection within each experiment or panel, or across
        # all experiments and panels, or just by images
        self._separate_experiments = separate_experiments
        self._separate_panels = separate_panels
        self._separate_images = separate_images

        # block width for splitting scans over phi, or None for no split
        self._block_width = block_width

        # the number of rejections
        self.nreject = 0

        self.nproc = nproc

        return

    def get_block_width(self, exp_id=None):
        if exp_id is None:
            return self._block_width
        else:
            try:
                bw = self._block_width[exp_id]
            except TypeError:
                bw = self._block_width
            return bw

    def set_block_width(self, block_width):
        """Set the block width for outlier detection in degrees. This can be either
        a single value or a list with one value per experiment. None is accepted
        to mean that the dataset will not be split into blocks."""
        self._block_width = block_width

    def _detect_outliers(self, cols):
        """Perform outlier detection using the input cols and return a flex.bool
        indicating which rows in the cols are considered outlying. cols should be
        a list of flex arrays of equal lengths"""

        # to be implemented by derived classes
        raise NotImplementedError()

    def __call__(self, reflections):
        """Identify outliers in the input and set the centroid_outlier flag.
        Return True if any outliers were detected, otherwise False"""

        logger.info(
            "Detecting centroid outliers using the %s algorithm", type(self).__name__
        )

        # check the columns are present
        for col in self._cols:
            assert col in reflections

        sel = reflections.get_flags(reflections.flags.predicted)
        all_data = reflections.select(sel)
        all_data_indices = sel.iselection()
        nexp = flex.max(all_data["id"]) + 1

        jobs = []
        if self._separate_experiments:
            # split the data set by experiment id
            for iexp in range(nexp):
                sel = all_data["id"] == iexp
                job = {
                    "id": iexp,
                    "panel": "all",
                    "data": all_data.select(sel),
                    "indices": all_data_indices.select(sel),
                }
                jobs.append(job)
        else:
            # keep the whole dataset across all experiment ids
            job = {
                "id": "all",
                "panel": "all",
                "data": all_data,
                "indices": all_data_indices,
            }
            jobs.append(job)

        jobs2 = []
        if self._separate_panels:
            # split further by panel id
            for job in jobs:
                data = job["data"]
                iexp = job["id"]
                indices = job["indices"]
                for ipanel in range(flex.max(data["panel"]) + 1):
                    sel = data["panel"] == ipanel
                    job = {
                        "id": iexp,
                        "panel": ipanel,
                        "data": data.select(sel),
                        "indices": indices.select(sel),
                    }
                    jobs2.append(job)
        else:
            # keep the splits as they are
            jobs2 = jobs

        jobs3 = []
        if self.get_block_width() is not None:
            # split into equal-sized phi ranges
            for job in jobs2:
                data = job["data"]
                iexp = job["id"]
                ipanel = job["panel"]
                indices = job["indices"]
                phi = data["xyzobs.mm.value"].parts()[2]
                if len(phi) == 0:  # detect no data in the job
                    jobs3.append(job)
                    continue
                phi_low = flex.min(phi)
                phi_range = flex.max(phi) - phi_low
                if phi_range == 0.0:  # detect stills and do not split
                    jobs3.append(job)
                    continue
                bw = self.get_block_width(iexp)
                if bw is None:  # detect no split for this experiment
                    jobs3.append(job)
                    continue
                nblocks = int(round(math.degrees(phi_range / bw)))
                nblocks = max(1, nblocks)
                real_width = phi_range / nblocks
                block_end = 0.0
                for iblock in range(nblocks - 1):  # all except the last block
                    block_start = iblock * real_width
                    block_end = (iblock + 1) * real_width
                    sel = (phi >= (phi_low + block_start)) & (
                        phi < (phi_low + block_end)
                    )
                    job = {
                        "id": iexp,
                        "panel": ipanel,
                        "data": data.select(sel),
                        "indices": indices.select(sel),
                        "phi_start": math.degrees(phi_low + block_start),
                        "phi_end": math.degrees(phi_low + block_end),
                    }
                    jobs3.append(job)
                # now last block
                sel = phi >= (phi_low + block_end)
                job = {
                    "id": iexp,
                    "panel": ipanel,
                    "data": data.select(sel),
                    "indices": indices.select(sel),
                    "phi_start": math.degrees(phi_low + block_end),
                    "phi_end": math.degrees(phi_low + phi_range),
                }
                jobs3.append(job)
        elif self._separate_images:
            for job in jobs2:
                data = job["data"]
                iexp = job["id"]
                ipanel = job["panel"]
                indices = job["indices"]
                images = data["xyzobs.px.value"].parts()[2].iround()
                if len(images) == 0:  # detect no data in the job
                    jobs3.append(job)
                    continue
                images_low = flex.min(images)
                images_high = flex.max(images)
                for iblock in range(images_low, images_high + 1):
                    sel = images == iblock
                    image_refs = data.select(sel)
                    phi = image_refs["xyzobs.mm.value"].parts()[2]
                    job = {
                        "id": iexp,
                        "panel": ipanel,
                        "data": image_refs,
                        "indices": indices.select(sel),
                        "phi_start": flex.min(phi),
                        "phi_end": flex.max(phi),
                    }
                    jobs3.append(job)
        else:
            # keep the splits as they are
            jobs3 = jobs2

        # Work out the format of the jobs table
        header = ["Job"]
        if self._separate_experiments:
            header.append("Exp\nid")
        if self._separate_panels:
            header.append("Panel\nid")
        if self.get_block_width() is not None:
            header.append("Block range\n(deg)")
        header.extend(["Nref", "Nout", "%out"])
        rows = []

        # Now loop over the lowest level of splits and run outlier detection
        if self.nproc > 1:
            with concurrent.futures.ProcessPoolExecutor(max_workers=self.nproc) as pool:
                outlier_detection_runs = [
                    pool.submit(self._run_job, job, i) for i, job in enumerate(jobs3)
                ]
            outlier_detection_runs = [e.result() for e in outlier_detection_runs]
        else:
            # For nproc=1 keep the jobs in the main process
            outlier_detection_runs = [
                self._run_job(job, i) for i, job in enumerate(jobs3)
            ]
        # Copy results back into the job dict and report any messages
        for result in outlier_detection_runs:
            job = jobs3[result["index"]]
            job["ioutliers"] = result["ioutliers"]
            if result["message"]:
                logger.debug(result["message"])

        # loop over the completed jobs
        for i, job in enumerate(jobs3):

            iexp = job["id"]
            ipanel = job["panel"]
            nref = len(job["indices"])
            ioutliers = job["ioutliers"]

            # set the centroid_outlier flag in the original reflection table
            nout = len(ioutliers)
            if nout > 0:
                reflections.set_flags(ioutliers, reflections.flags.centroid_outlier)
                self.nreject += nout

            # Add job data to the table
            row = [str(i + 1)]
            if self._separate_experiments:
                row.append(str(iexp))
            if self._separate_panels:
                row.append(str(ipanel))
            if self.get_block_width() is not None:
                try:
                    row.append("{phi_start:.2f} - {phi_end:.2f}".format(**job))
                except KeyError:
                    row.append(f"{0.0:.2f} - {0.0:.2f}")
            if nref == 0:
                p100 = 0
            else:
                p100 = nout / nref * 100.0
                if p100 > 30.0:
                    msg = (
                        "{:3.1f}% of reflections were flagged as outliers from job"
                        " {}"
                    ).format(p100, i + 1)
                    logger.debug(msg)
            row.extend([str(nref), str(nout), f"{p100:3.1f}"])
            rows.append(row)

        if self.nreject == 0:
            return False
        logger.info(f"{self.nreject} reflections have been flagged as outliers")
        logger.debug("Outlier rejections per job:")
        logger.debug(tabulate(rows, header))

        return True

    def _run_job(self, job, i):
        data = job["data"]
        indices = job["indices"]
        nref = len(indices)

        msg = None
        if nref >= self._min_num_obs:

            # get the subset of data as a list of columns
            cols = [data[col] for col in self._cols]

            # determine the position of outliers on this sub-dataset
            outliers = self._detect_outliers(cols)

            # get positions of outliers from the original matches
            ioutliers = indices.select(outliers)

        elif nref > 0:
            # too few reflections in the job
            msg = "For job {}, fewer than {} reflections are present.".format(
                i + 1, self._min_num_obs
            )
            msg += " All reflections flagged as possible outliers."
            ioutliers = indices

        else:
            # no reflections in the job
            ioutliers = indices

        return {"index": i, "message": msg, "ioutliers": ioutliers}


# The phil scope for outlier rejection
phil_str = """
outlier
  .help = "Outlier rejection after initial reflection prediction."
{
  algorithm = null *auto mcd tukey sauter_poon
    .help = "Outlier rejection algorithm. If auto is selected, the algorithm is"
            "chosen automatically."
    .type = choice
    .short_caption = "Outlier rejection algorithm"

  nproc = 1
    .help = "Number of processes over which to split outlier identification."
            "If set to Auto, DIALS will choose automatically."
    .type = int(value_min=1)
    .expert_level = 1

  minimum_number_of_reflections = 20
    .help = "The minimum number of input observations per outlier rejection"
            "job below which all reflections in the job will be rejected as"
            "potential outliers."
    .type = int(value_min=0)
    .expert_level = 1

  separate_experiments = True
    .help = "If true, outlier rejection will be performed on each experiment"
            "separately. Otherwise, the data from all experiments will be"
            "combined for outlier rejection."
    .type = bool
    .expert_level = 1

  separate_panels = False
    .help = "Perform outlier rejection separately for each panel of a multi-"
            "panel detector model. Otherwise data from across all panels will"
            "be combined for outlier rejection."
    .type = bool
    .expert_level = 1

  separate_blocks = True
    .help = "If true, for scans outlier rejection will be performed separately"
            "in equal-width blocks of phi, controlled by the parameter"
            "outlier.block_width."
    .type = bool
    .expert_level = 1

  block_width = Auto
    .help = "If separate_blocks, a scan will be divided into equal-sized blocks"
            "with width (in degrees) close to this value for outlier rejection."
            "If Auto, a width of at least 18 degrees will be determined,"
            "such that each block contains enough reflections to perform"
            "outlier rejection."
    .type = float(value_min=1.0)
    .expert_level = 1

  separate_images = False
    .help = "If true, every image will be treated separately for outlier"
            "rejection. It is a special case that will override both"
            "separate_experiments and separate_blocks, and will set these"
            "to False if required."
    .type = bool
    .expert_level = 2

  tukey
    .help = "Options for the tukey outlier rejector"
    .expert_level = 1
  {
    iqr_multiplier = 1.5
      .help = "The IQR multiplier used to detect outliers. A value of 1.5"
              "gives Tukey's rule for outlier detection"
      .type = float(value_min = 0.)
  }

  mcd
    .help = "Options for the mcd outlier rejector, which uses an algorithm"
            "based on FAST-MCD by Rousseeuw and van Driessen. See"
            "doi.org/10.1080/00401706.1999.10485670."
    .expert_level = 1
  {
     alpha = 0.5
       .help = "Decimal fraction controlling the size of subsets over which the"
               "covariance matrix determinant is minimised."
       .type = float(value_min = 0., value_max = 1.0)

     max_n_groups=5
       .help = "The maximum number of groups to split the dataset into if the"
               "dataset is 'large' (more observations than twice the"
               "min_group_size)."
       .type = int(value_min = 1)

     min_group_size=300
       .help = "The smallest sub-dataset size when splitting the dataset into"
               "a number of groups, maximally max_n_groups."
       .type = int(value_min = 100)

     n_trials=500
       .help = "The number of samples used for initial estimates to seed the"
               "search within each sub-dataset."
       .type = int(value_min = 1)

     k1=2
       .help = "The number of concentration steps to take after initial estimates."
       .type = int(value_min = 1)

     k2=2
       .help = "If the dataset is 'large', the number of concentration steps to"
               "take after applying the best subset estimates to the merged"
               "group."
       .type = int(value_min = 1)

     k3=100
       .help = "If the dataset is 'small', the number of concentration steps to"
               "take after selecting the best of the initial estimates, applied"
               "to the whole dataset."
       .type = int(value_min = 1)

     threshold_probability=0.975
       .help = "Quantile probability from the Chi-squared distribution with"
               "number of degrees of freedom equal to the number of dimensions"
               "of the data data (e.g. 3 for X, Y and Phi residuals)."
               "Observations whose robust Mahalanobis distances are larger than"
               "the obtained quantile will be flagged as outliers."
       .type = float(value_min = 0., value_max = 1.0)
  }

  sauter_poon
    .help = "Options for the outlier rejector described in Sauter & Poon (2010)"
            "(https://doi.org/10.1107/S0021889810010782)"
    .expert_level = 1
  {
    px_sz = Auto
      .help = "X, Y pixel size in mm. If Auto, this will be taken from"
              "the first panel of the first experiment."
      .type = floats(size = 2, value_min = 0.001)

    verbose = False
      .help = "Verbose output."
      .type = bool
      .multiple = False

    pdf=None
      .help = "Output file name for making graphs of |dr| vs spot number and dy vs dx."
      .type = str
      .multiple = False
  }
}
"""

phil_scope = parse(phil_str)


class CentroidOutlierFactory:
    @staticmethod
    def from_parameters_and_colnames(params, colnames):

        # id the relevant scope for the requested method
        method = params.outlier.algorithm
        if method == "null":
            return None
        elif method == "tukey":
            from dials.algorithms.refinement.outlier_detection.tukey import (
                Tukey as outlier_detector,
            )

            algo_params = params.outlier.tukey
        elif method == "mcd":
            from dials.algorithms.refinement.outlier_detection.mcd import (
                MCD as outlier_detector,
            )

            algo_params = params.outlier.mcd
        elif method == "sauter_poon":
            from dials.algorithms.refinement.outlier_detection.sauter_poon import (
                SauterPoon as outlier_detector,
            )

            algo_params = params.outlier.sauter_poon
        else:
            raise RuntimeError("outlier.algorithm not recognised")

        # construct kwargs from the algo_params scope
        kwargs = {
            k: v for k, v in algo_params.__dict__.items() if not k.startswith("_")
        }

        # Special case where we ignore experiment and blocks and just look
        # for outliers on each image independently
        if params.outlier.separate_images:
            if params.outlier.separate_blocks:
                logger.info("Resetting separate_blocks=False")
                params.outlier.separate_blocks = False
            if params.outlier.separate_experiments:
                logger.info("Resetting separate_experiments=False")
                params.outlier.separate_experiments = False

        if not params.outlier.separate_blocks:
            params.outlier.block_width = None

        if params.outlier.nproc is libtbx.Auto:
            params.outlier.nproc = available_cores()
            logger.info("Setting outlier.nproc={}".format(params.outlier.nproc))

        od = outlier_detector(
            cols=colnames,
            min_num_obs=params.outlier.minimum_number_of_reflections,
            separate_experiments=params.outlier.separate_experiments,
            separate_panels=params.outlier.separate_panels,
            separate_images=params.outlier.separate_images,
            block_width=params.outlier.block_width,
            nproc=params.outlier.nproc,
            **kwargs,
        )
        return od


if __name__ == "__main__":

    # test construction
    params = phil_scope.extract()
    params.outlier.algorithm = "tukey"
    print(CentroidOutlierFactory.from_parameters_and_colnames(params, [1, 2, 3]))
    params.outlier.algorithm = "mcd"
    print(CentroidOutlierFactory.from_parameters_and_colnames(params, [1, 2, 3]))



from __future__ import annotations

import random

import pytest

from cctbx import sgtbx
from scitbx import matrix
from scitbx.math import euler_angles_as_matrix

random.seed(42)


def random_rotation(angle_min=0, angle_max=360):
    angles = [random.uniform(angle_min, angle_max) for i in range(3)]
    print("Rotation: ", angles)
    return euler_angles_as_matrix(angles, deg=True)


@pytest.fixture(params=["P2", "P3", "P6", "R3:h", "I23"])
def setup_rlp(request):
    random.seed(42)  # guaranteed to be random
    space_group = request.param
    # setup symmetry information
    sgi = sgtbx.space_group_info(symbol=space_group)
    cs = sgi.any_compatible_crystal_symmetry(volume=10000)
    cs = cs.best_cell()
    cs = cs.minimum_cell()
    uc = cs.unit_cell()

    B = matrix.sqr(uc.fractionalization_matrix()).transpose()
    U = random_rotation()
    A = U * B

    ms = cs.build_miller_set(d_min=2, anomalous_flag=True).expand_to_p1()
    rlp = A.elems * ms.indices().as_vec3_double()

    d = {}
    d["crystal_symmetry"] = cs
    d["rlp"] = rlp
    return d



from __future__ import annotations


def test_load_and_dump():
    from dials.algorithms.profile_model.gaussian_rs import Model

    d1 = {"__id__": "gaussian_rs", "n_sigma": 3, "sigma_b": 1, "sigma_m": 2}
    d2 = {"__id__": "gaussian_rs", "n_sigma": 2, "sigma_b": 4, "sigma_m": 5}
    model1 = Model.from_dict(d1)
    model2 = Model.from_dict(d2)
    assert model1.n_sigma() == 3
    assert model1.sigma_b() == 1
    assert model1.sigma_m() == 2
    assert model2.n_sigma() == 2
    assert model2.sigma_b() == 4
    assert model2.sigma_m() == 5



from __future__ import annotations

import sys

import iotbx.phil

import dials.util

help_message = """

Convert data which can be read by DIALS, given a experiment list, to CBF format -
with ersatz miniCBF header. Can be used e.g. with HDF5 format data.

Examples::

  dials.convert_to_cbf models.expt prefix=data_as_cbf
"""

phil_scope = iotbx.phil.parse(
    """\
output {
  template = as_cbf_%04d.cbf
    .type = path
}
""",
    process_includes=True,
)


def convert_to_cbf(imageset, template):
    from dxtbx.format.FormatCBFMini import FormatCBFMini

    for i in range(len(imageset)):
        print(template % (i + 1))
        FormatCBFMini.as_file(
            imageset.get_detector(),
            imageset.get_beam(),
            imageset.get_goniometer(),
            imageset.get_scan()[i],
            imageset.get_raw_data(i)[0],
            template % (i + 1),
        )


@dials.util.show_mail_handle_errors()
def run(args=None):
    from dials.util.options import ArgumentParser, flatten_experiments

    usage = "dials.convert_to_cbf [options] models.expt"

    parser = ArgumentParser(
        usage=usage,
        phil=phil_scope,
        read_experiments=True,
        read_experiments_from_images=True,
        epilog=help_message,
    )

    params, options, args = parser.parse_args(
        args, show_diff_phil=True, return_unhandled=True
    )

    template = params.output.template
    experiments = flatten_experiments(params.input.experiments)

    if len(experiments) == 0:
        parser.print_help()
        return

    if len(experiments) > 1:
        sys.exit("Only one experiment can be processed at a time")
    else:
        imagesets = experiments.imagesets()
        assert len(imagesets) == 1, len(imagesets)
        imageset = imagesets[0]

    convert_to_cbf(imageset, template)


if __name__ == "__main__":
    run()



from __future__ import annotations

from math import exp

import pytest

from dials.algorithms.scaling.model.components.smooth_scale_components import (
    GaussianSmoother2D,
    GaussianSmoother3D,
)
from dials.array_family import flex


def test_2DGaussianSmoother():
    """Test the behaviour of the 2D Gaussian Smoother"""
    GS2D = GaussianSmoother2D([0, 1], 1, [0, 4], 4)
    assert GS2D.num_x_values() == 2
    assert GS2D.num_y_values() == 6
    assert GS2D.num_x_average() == 2
    assert GS2D.num_y_average() == 3
    parameters = flex.double(
        [1.0, 1.0, 2.0, 2.0, 3.0, 3.0, 4.0, 4.0, 5.0, 5.0, 6.0, 6.0]
    )
    x = flex.double([0.5, 0.25])
    y = flex.double([1.0, 0.75])
    value, weight, sumw = GS2D.multi_value_weight(x, y, parameters)
    assert weight.non_zeroes == 6 * 2
    assert GS2D.sigma() == 0.65
    # Calculate tand verify the expected value at the first position.
    expected_sumw = (2.0 * exp(-2.5 / (0.65**2))) + 4.0 * exp(-0.5 / (0.65**2))
    assert expected_sumw == pytest.approx(sumw[0])
    expected_value_numerator = (10.0 * exp(-0.5 / (0.65**2))) + (
        2.0 * exp(-2.5 / (0.65**2))
    )
    assert value[0] == pytest.approx(expected_value_numerator / expected_sumw)

    # Do the same calculation but calling the single value_weight function.
    x = 0.5
    y = 1.0
    value, weight, sumw = GS2D.value_weight(x, y, parameters)
    assert weight.non_zeroes == 6
    assert value == pytest.approx(expected_value_numerator / expected_sumw)

    # Test again for another small number of params
    GS2D = GaussianSmoother2D([0, 2], 2, [0, 4], 4)
    assert GS2D.num_x_values() == 3
    assert GS2D.num_y_values() == 6
    assert GS2D.num_x_average() == 3
    assert GS2D.num_y_average() == 3

    # Test again for another number of params
    GS2D = GaussianSmoother2D([0, 4], 4, [0, 5], 5)
    assert GS2D.num_x_values() == 6
    assert GS2D.num_y_values() == 7
    assert GS2D.num_x_average() == 3
    assert GS2D.num_y_average() == 3


def test_3DGaussianSmoother():
    """Test the behaviour of the 2D Gaussian Smoother"""
    GS3D = GaussianSmoother3D([0, 1], 1, [0, 3], 3, [0, 1], 1)
    assert GS3D.num_x_values() == 2
    assert GS3D.num_y_values() == 5
    assert GS3D.num_z_values() == 2
    assert GS3D.num_x_average() == 2
    assert GS3D.num_y_average() == 3
    assert GS3D.num_z_average() == 2
    parameters = flex.double([1.0, 1.0, 2.0, 2.0, 3.0, 3.0, 4.0, 4.0, 5.0, 5.0] * 2)
    x = flex.double([0.5, 0.25])
    y = flex.double([1.0, 0.75])
    z = flex.double([0.5, 0.25])
    value, weight, sumw = GS3D.multi_value_weight(x, y, z, parameters)
    assert weight.non_zeroes == 6 * 2 * 2
    assert GS3D.sigma() == 0.65

    # Calculate tand verify the expected value at the first position.
    expected_sumw = (4.0 * exp(-2.75 / (0.65**2))) + (8.0 * exp(-0.75 / (0.65**2)))
    assert expected_sumw == pytest.approx(sumw[0])
    expected_value_numerator = (20.0 * exp(-0.75 / (0.65**2))) + (
        4.0 * exp(-2.75 / (0.65**2))
    )
    assert value[0] == pytest.approx(expected_value_numerator / expected_sumw)

    # Do the same calculation but calling the single value_weight function.
    x = 0.5
    y = 1.0
    z = 0.5
    value, weight, sumw = GS3D.value_weight(x, y, z, parameters)
    assert weight.non_zeroes == 6 * 2
    assert value == pytest.approx(expected_value_numerator / expected_sumw)

    # Test again for another number of params
    GS3D = GaussianSmoother3D([0, 1], 1, [0, 2], 2, [0, 3], 3)
    assert GS3D.num_x_values() == 2
    assert GS3D.num_y_values() == 3
    assert GS3D.num_z_values() == 5
    assert GS3D.num_x_average() == 2
    assert GS3D.num_y_average() == 3
    assert GS3D.num_y_average() == 3



from __future__ import annotations

from dials.util import parse_htcondor_job_classad


def test_parse_htcondor_job_classad(tmp_path):
    machine_ad = tmp_path / "condor_job_ad"
    machine_ad.write_text(
        """\
NumShadowStarts = 1
MemoryProvisioned = 4096
CpusProvisioned = 8
TransferIn = false
"""
    )
    class_ad = parse_htcondor_job_classad(machine_ad)
    assert class_ad.cpus_provisioned == 8
    assert class_ad.memory_provisioned == 4096



from __future__ import annotations

from dials.algorithms.symmetry import origin
from dials.array_family import flex


def test_origin_offset_miller_indices():
    mi = flex.miller_index(
        [(h, k, l) for h in range(5) for k in range(5) for l in range(5)]
    )
    omi = origin.offset_miller_indices(mi, (0, 1, -1))
    ref = flex.miller_index(
        [(h, k, l) for h in range(5) for k in range(1, 6) for l in range(-1, 4)]
    )

    assert ref == omi



from __future__ import annotations


def build_prediction_matrix(hkl, mhkl, phkl, hmkl, hpkl, hkml, hkpl, tst=False):
    """Build a prediction matrix around reflection hkl, coomputing dx / dh etc."""

    from scitbx import matrix

    refl = hkl, mhkl, phkl, hmkl, hpkl, hkml, hkpl

    x = [r.image_coord_px[0] for r in refl]
    y = [r.image_coord_px[1] for r in refl]
    z = [r.frame_number for r in refl]

    d = 0.5 * matrix.sqr(
        (
            x[2] - x[1],
            x[4] - x[3],
            x[6] - x[5],
            y[2] - y[1],
            y[4] - y[3],
            y[6] - y[5],
            z[2] - z[1],
            z[4] - z[3],
            z[6] - z[5],
        )
    )

    if not tst:
        return d

    # now assess how accurate this d matrix is

    for j, r in enumerate(refl[1:]):
        dh = matrix.col(
            (
                r.miller_index[0] - hkl.miller_index[0],
                r.miller_index[1] - hkl.miller_index[1],
                r.miller_index[2] - hkl.miller_index[2],
            )
        )
        dxyz = d * dh
        dx = x[0] + dxyz[0] - x[j + 1]
        dy = y[0] + dxyz[1] - y[j + 1]
        dz = z[0] + dxyz[2] - z[j + 1]

        print(j, dx, dy, dz)

    return d



from __future__ import annotations


class SimpleCentroidExt:
    """An extension class implementing a simple centroid algorithm."""

    name = "simple"

    default = True

    def __init__(self, params, experiments):
        """Initialise the algorithm.

        :param params: The input phil parameters
        :param experiments: The experiment list
        """
        self.experiments = experiments

    def compute_centroid(self, reflections, image_volume=None):
        """
        Compute the centroid.

        :param reflections: The list of reflections
        """
        import dials.algorithms.centroid.simple

        return dials.algorithms.centroid.simple.centroid(
            self.experiments, reflections, image_volume=image_volume
        )



from __future__ import annotations

from scitbx.array_family import flex

from dials.algorithms.clustering import plots
from dials.util.observer import Observer, singleton


def uc_params_from_experiments(experiments):
    uc_params = [flex.double() for i in range(6)]
    for expt in experiments:
        uc = expt.crystal.get_unit_cell()
        for i in range(6):
            uc_params[i].append(uc.parameters()[i])
    return uc_params


@singleton
class UnitCellAnalysisObserver(Observer):
    """
    Observer to record unit cell clustering data and make plots.
    """

    def update(self, script):
        """Update the data in the observer."""
        try:
            self.data["dendrogram"] = script.unit_cell_dendrogram
        except AttributeError:
            pass
        self.data["experiments"] = script._experiments

    def make_plots(self):
        """Generate plots of the unit cell clustering."""

        uc_params = uc_params_from_experiments(self.data["experiments"])
        d = plots.plot_uc_histograms(uc_params)

        if "dendrogram" in self.data:
            d["uc_clustering"] = plots.scipy_dendrogram_to_plotly_json(
                self.data["dendrogram"],
                title="Unit cell clustering",
                xtitle="Dataset",
                ytitle=r"Distance (Å<sup>2</sup>)",
                help="""\
The results of single-linkage hierarchical clustering on the unit cell parameters using
the Andrews-Bernstein NCDist distance metric (Andrews & Bernstein, 2014). The height at
which two clusters are merged in the dendrogram is a measure of the similarity between
the unit cells in each cluster. A larger separation between two clusters may be
indicative of a higher degree of non-isomorphism between the clusters. Conversely, a
small separation between two clusters suggests that their unit cell parameters are
relatively isomorphous.
""",
            )

        graphs = {"unit_cell_graphs": d}

        return graphs



from __future__ import annotations

import binascii
import sys

import iotbx.phil
from dxtbx.ext import compress, uncompress

import dials.util

help_message = """

This program can be used to merge a given number of consecutive cbf files into
a smaller number of images For example, running dials.merge_cbf on a experiments
with 100 images, using the default value of merge_n_images=2, will output 50
summed images, with every consecutive pair of images being summed into a single
output image. Currently only cbf format images are supported as input.

Examples::

  dials.merge_cbf image_*.cbf

  dials.merge_cbf image_*.cbf merge_n_images=10
"""

phil_scope = iotbx.phil.parse(
    """\
merge_n_images = 2
  .type = int(value_min=1)
  .help = "Number of input images to average into a single output image"

get_raw_data_from_imageset = True
  .type = bool
  .help = "By default the raw data is read via the imageset. This limits use"
          "to single panel detectors where the format class does not make"
          "modifications to the array size in the file. Set this option to"
          "false in order to bypass the imageset and read the data as-is from"
          "the CBF file"
  .expert_level = 2

output {
  image_prefix = sum_
    .type = path
}
""",
    process_includes=True,
)


def get_raw_data_from_file(imageset, i):
    """Read cbf directly to access the raw data array rather than
    through the imageset, in order to work for multi-panel detectors and other
    situations where the format class modifies the raw array"""
    file_name = imageset.get_image_identifier(i)
    with open(file_name, "rb") as cbf:
        data = cbf.read()
    start_tag = binascii.unhexlify("0c1a04d5")
    data_offset = data.find(start_tag) + 4
    cbf_header = data[: data_offset - 4].decode("latin-1")
    fast = slow = length = 0
    for record in cbf_header.split("\n"):
        if "X-Binary-Size-Fastest-Dimension" in record:
            fast = int(record.split()[-1])
        elif "X-Binary-Size-Second-Dimension" in record:
            slow = int(record.split()[-1])
        elif "X-Binary-Size:" in record:
            length = int(record.split()[-1])
    values = uncompress(
        packed=data[data_offset : data_offset + length], fast=fast, slow=slow
    )
    return (values,)


def merge_cbf(imageset, n_images, out_prefix="sum_", get_raw_data_from_imageset=True):
    from dxtbx.format.FormatCBF import FormatCBF

    assert issubclass(
        imageset.get_format_class(), FormatCBF
    ), "Only CBF format images supported"

    assert len(imageset) >= n_images

    n_output_images = len(imageset) // n_images

    n_digits = len(str(n_output_images))

    for i_out in range(n_output_images):
        data_out = None

        for j in range(n_images):

            i_in = (i_out * n_images) + j

            if get_raw_data_from_imageset:
                data_in = imageset.get_raw_data(i_in)
            else:
                data_in = get_raw_data_from_file(imageset, i_in)

            assert len(data_in) == 1
            data_in = data_in[0]
            if data_out is None:
                data_out = data_in
            else:
                # FIXME only add pixels to this which are > 0; image pixels < 0
                # are meaningful and should be preserved;
                # Achieved by setting -ve values here to 0 before +=
                # This assumes that -ve values are constant over all images
                data_special = data_in < 0
                data_in.set_selected(data_special, 0)
                data_out += data_in

        out_image = "{prefix}{number:0{digits}d}.cbf".format(
            prefix=out_prefix, number=i_out + 1, digits=n_digits
        )

        start_tag = binascii.unhexlify("0c1a04d5")

        with open(imageset.get_path(i_out * n_images), "rb") as fh:
            data = fh.read()
        data_offset = data.find(start_tag)
        cbf_header = data[:data_offset].decode("latin-1")

        new_header = []
        compressed = compress(data_out)

        old_size = 0

        for record in cbf_header.split("\n")[:-1]:
            rsplit = record.split(" ")
            if "X-Binary-Size:" in record:
                old_size = int(record.split()[-1])
                new_header.append(f"X-Binary-Size: {len(compressed)}\r\n")
            elif "Content-MD5" in record:
                pass
            elif len(rsplit) > 3 and rsplit[1] in {
                "Exposure_time",
                "Angle_increment",
                "Exposure_period",
                "Count_cutoff",
                "Phi_increment",
                "Omega_increment",
                "Chi_increment",
            }:

                if rsplit[1] == "Count_cutoff":  # needs to be an integer
                    new_header.append(
                        "%s\n"
                        % " ".join(
                            rsplit[:2]
                            + ["%d" % (n_images * int(rsplit[2]))]
                            + rsplit[3:]
                        )
                    )
                else:
                    new_header.append(
                        "%s\n"
                        % " ".join(
                            rsplit[:2]
                            + ["%f" % (n_images * float(rsplit[2]))]
                            + rsplit[3:]
                        )
                    )

            else:
                new_header.append(f"{record}\n")

        loop_lines = [
            n for n, record in enumerate(new_header) if record.startswith("loop_")
        ]
        multiply_fields = {
            "_diffrn_scan_axis.angle_range",
            "_diffrn_scan_axis.angle_increment",
            "_diffrn_scan_axis.displacement_range",
            "_diffrn_scan_axis.displacement_increment",
            "_diffrn_scan_frame.integration_time",
            "_diffrn_scan_frame.exposure_time",
            "_array_intensities.overload",
        }
        for loop_start in loop_lines:
            n = loop_start
            modifiers = []
            while True:
                n = n + 1
                line = new_header[n].strip()
                if line in {"", ";"}:  # end of loop
                    break
                elif line.startswith("_"):  # loop header
                    if line in multiply_fields:
                        modifiers.append(n_images)
                    else:
                        modifiers.append(None)
                elif any(modifiers):  # loop body
                    # NOTE: This can break when fields are modified in loops with
                    #   'Strings with spaces, as they are seen as multiple columns, or with'
                    #   _multiple _columns _defined _on _same _line _they _are _seen _as _one _column
                    new_line = [
                        element
                        if modifier is None
                        else f"{float(element) * modifier:f}"
                        for modifier, element in zip(modifiers, line.split())
                    ]
                    new_header[n] = f"{' '.join(new_line)}\r\n"

        tailer = data[data_offset + 4 + old_size :]

        with open(out_image, "wb") as f:
            f.write("".join(new_header).encode("latin-1"))
            f.write(start_tag)
            f.write(compressed)
            f.write(tailer)
        print(f"{out_image} written")


@dials.util.show_mail_handle_errors()
def run(args=None):
    from dials.util.options import ArgumentParser, flatten_experiments

    usage = "dials.merge_cbf [options] image_*.cbf"

    parser = ArgumentParser(
        usage=usage,
        phil=phil_scope,
        read_experiments=True,
        read_experiments_from_images=True,
        epilog=help_message,
    )

    params, options, args = parser.parse_args(
        args, show_diff_phil=True, return_unhandled=True
    )

    n_images = params.merge_n_images
    out_prefix = params.output.image_prefix
    experiments = flatten_experiments(params.input.experiments)

    if len(experiments) == 0:
        parser.print_help()
        return

    if len(experiments) > 1:
        sys.exit("Only one experiment can be processed at a time")
    else:
        imagesets = experiments.imagesets()
        assert len(imagesets) == 1
        imageset = imagesets[0]

    merge_cbf(
        imageset,
        n_images,
        out_prefix=out_prefix,
        get_raw_data_from_imageset=params.get_raw_data_from_imageset,
    )


if __name__ == "__main__":
    run()



from __future__ import annotations

import copy
from collections import namedtuple
from math import exp
from random import randint, uniform

import numpy as np
import pytest
from numpy.linalg import norm

from scitbx import matrix

from dials.algorithms.profile_model.ellipsoid.model import (
    Simple1ProfileModel,
    Simple6ProfileModel,
    compute_change_of_basis_operation,
)
from dials.algorithms.profile_model.ellipsoid.parameterisation import (  # Angular2MosaicityParameterisation,; Angular4MosaicityParameterisation,; WavelengthSpreadParameterisation,
    ModelState,
    Simple1MosaicityParameterisation,
    Simple6MosaicityParameterisation,
)
from dials.algorithms.profile_model.ellipsoid.refiner import (
    Refiner,
    RefinerData,
    ReflectionLikelihood,
    rotate_mat3_double,
    rotate_vec3_double,
)
from dials.algorithms.refinement.parameterisation.crystal_parameters import (
    CrystalOrientationParameterisation,
    CrystalUnitCellParameterisation,
)
from dials.array_family import flex


def first_derivative(func, x, h):
    return (-func(x + 2 * h) + 8 * func(x + h) - 8 * func(x - h) + func(x - 2 * h)) / (
        12 * h
    )


def generate_data(experiments, reflections):

    from random import seed

    seed(0)

    index = randint(0, len(reflections))

    h = reflections[index]["miller_index"]

    s0 = matrix.col(experiments[0].beam.get_s0())

    U_param = CrystalOrientationParameterisation(experiments[0].crystal)
    B_param = CrystalUnitCellParameterisation(experiments[0].crystal)

    U = matrix.sqr(experiments[0].crystal.get_U())
    B = matrix.sqr(experiments[0].crystal.get_B())
    r = U * B * matrix.col(h)
    s2 = s0 + r
    mobs = (
        s2 + matrix.col((uniform(0, 1e-3), uniform(0, 1e-3), uniform(0, 1e-3)))
    ).normalize() * s0.length()
    mobs = np.array([uniform(0, 1e-3), uniform(0, 1e-3)])
    sp = s2.normalize() * s0.length()

    b1, b2, b3, b4, b5, b6 = (
        uniform(1e-3, 3e-3),
        uniform(0.0, 1e-3),
        uniform(1e-3, 3e-3),
        uniform(0.0, 1e-3),
        uniform(0.0, 1e-3),
        uniform(1e-3, 3e-3),
    )

    S_param = (b1, b2, b3, b4, b5, b6)
    L_param = (uniform(1e-3, 2e-3),)
    ctot = randint(100, 1000)

    T = np.array(
        (uniform(1e-3, 2e-3), 0, uniform(1e-6, 2e-6), uniform(1e-3, 2e-3))
    ).reshape(2, 2)
    Sobs = np.matmul(T, T.T)

    params = [S_param, U_param, B_param, L_param]

    return params, s0, sp, h, ctot, mobs, Sobs


@pytest.fixture
def testdata(test_experiment):

    TestData = namedtuple(
        "TestData",
        [
            "experiment",
            "reflections",
            "models",
            "s0",
            "sp",
            "h",
            "ctot",
            "mobs",
            "Sobs",
        ],
    )

    experiments = [test_experiment]
    reflections = flex.reflection_table.from_predictions_multi(experiments)

    models, s0, sp, h, ctot, mobs, Sobs = generate_data(experiments, reflections)

    return TestData(
        experiment=experiments[0],
        reflections=reflections,
        models=models,
        s0=s0,
        sp=sp,
        h=h,
        ctot=ctot,
        mobs=mobs,
        Sobs=Sobs,
    )


@pytest.fixture
def refinerdata_testdata(testdata):

    experiment = testdata.experiment
    reflections = testdata.reflections

    panel = experiment.detector[0]
    s0_length = matrix.col(experiment.beam.get_s0()).length()
    reflections["bbox"] = flex.int6(len(reflections))
    reflections["xyzobs.px.value"] = flex.vec3_double(len(reflections))
    reflections["s2"] = reflections["s1"].each_normalize() * s0_length
    reflections["sp"] = flex.vec3_double(len(reflections))
    for i, (x, y, z) in enumerate(reflections["xyzcal.px"]):
        x0 = int(x) - 5
        x1 = int(x) + 5 + 1
        y0 = int(y) - 5
        y1 = int(y) + 5 + 1
        z0 = int(z)
        z1 = z0 + 1
        reflections["bbox"][i] = x0, x1, y0, y1, z0, z1
        reflections["xyzobs.px.value"][i] = (int(x) + 0.5, int(y) + 0.5, int(z) + 0.5)
        reflections["sp"][i] = (
            matrix.col(
                panel.get_pixel_lab_coord(reflections["xyzobs.px.value"][i][0:2])
            ).normalize()
            * s0_length
        )

    reflections["shoebox"] = flex.shoebox(
        reflections["panel"], reflections["bbox"], allocate=True
    )

    shoebox_data = flex.float(flex.grid(1, 11, 11))
    shoebox_mask = flex.int(flex.grid(1, 11, 11))
    for j in range(11):
        for i in range(11):
            shoebox_data[0, j, i] = (
                100
                * exp(-0.5 * (j - 5) ** 2 / 1**2)
                * exp(-0.5 * (i - 5) ** 2 / 1**2)
            )
            shoebox_mask[0, j, i] = 5
    for sbox in reflections["shoebox"]:
        sbox.data = shoebox_data
        sbox.mask = shoebox_mask

    return RefinerData.from_reflections(experiment, reflections)


def test_ConditionalDistribution(testdata):
    def check(
        mosaicity_parameterisation,
        wavelength_parameterisation,
        fix_mosaic_spread=False,
        fix_wavelength_spread=False,
        fix_unit_cell=False,
        fix_orientation=False,
    ):
        experiment = testdata.experiment
        models = testdata.models
        h = testdata.h
        s0 = np.array(testdata.s0).reshape(3, 1)
        sp = np.array(testdata.sp).reshape(3, 1)
        # ctot = testdata.ctot
        # mobs = testdata.mobs
        # Sobs = testdata.Sobs

        U_params = models[1].get_param_vals()
        B_params = models[2].get_param_vals()
        M_params = models[0][: mosaicity_parameterisation.num_parameters()]
        L_params = models[3]

        state = ModelState(
            experiment,
            mosaicity_parameterisation,
            wavelength_parameterisation,
            fix_mosaic_spread=fix_mosaic_spread,
            fix_wavelength_spread=fix_wavelength_spread,
            fix_unit_cell=fix_unit_cell,
            fix_orientation=fix_orientation,
        )
        state.U_params = np.array(U_params, dtype=np.float64)
        state.B_params = np.array(B_params, dtype=np.float64)
        state.M_params = np.array(M_params, dtype=np.float64)
        state.L_params = L_params

        def get_conditional(state):
            conditional = ReflectionLikelihood(
                state,
                s0,
                sp,
                h,
                0,
                np.array([0.0, 0.0]),
                np.array([0.0, 0.0, 0.0, 0.0]).reshape(2, 2),
            ).conditional
            return conditional

        conditional = get_conditional(state)

        step = 1e-6

        dm_dp = conditional.first_derivatives_of_mean()
        dS_dp = conditional.first_derivatives_of_sigma()
        parameters = state.active_parameters

        def compute_sigma(parameters):
            state.active_parameters = parameters
            return get_conditional(state).sigma()

        def compute_mean(parameters):
            state.active_parameters = parameters
            return get_conditional(state).mean()

        dm_num = []
        for i in range(len(parameters)):

            def f(x):
                p = copy.copy(parameters)
                p[i] = x
                return compute_mean(p)

            dm_num.append(first_derivative(f, parameters[i], step))

        for n, c in zip(dm_num, dm_dp):
            for nn, cc in zip(n, c):
                print(nn)
                print(cc)
                assert abs(nn - cc) < 1e-7
            # assert all(abs(nn - cc) < 1e-7 for nn, cc in zip(n, c))

        ds_num = []
        for i in range(len(parameters)):

            def f(x):
                p = copy.copy(parameters)
                p[i] = x
                return compute_sigma(p)

            ds_num.append(first_derivative(f, parameters[i], step))

        for n, c in zip(ds_num, dS_dp):
            for nn, cc in zip(n.flatten(), c.flatten()):
                print(nn)
                print(cc)
                assert abs(nn - cc) < 1e-7
            # assert all(abs(nn - cc) < 1e-7 for nn, cc in zip(n, c))

    S1 = Simple1MosaicityParameterisation()
    S6 = Simple6MosaicityParameterisation()

    check(S1, None, fix_wavelength_spread=True)
    check(S1, None, fix_wavelength_spread=True, fix_mosaic_spread=True)
    check(S1, None, fix_wavelength_spread=True, fix_unit_cell=True)
    check(S1, None, fix_wavelength_spread=True, fix_orientation=True)

    check(S6, None, fix_wavelength_spread=True)
    check(S6, None, fix_wavelength_spread=True, fix_mosaic_spread=True)
    check(S6, None, fix_wavelength_spread=True, fix_unit_cell=True)
    check(S6, None, fix_wavelength_spread=True, fix_orientation=True)


def test_rotate_vec3_double():

    vectors = np.array([[1, 1, 1]], dtype=np.float64).reshape(3, 1) / norm(
        np.array([1, 1, 1], dtype=np.float64)
    )

    v1 = np.array([0, 0, 1.0], dtype=np.float64).reshape(3, 1)
    R = compute_change_of_basis_operation(v1, vectors)

    rotated = rotate_vec3_double(R, vectors)

    assert rotated == pytest.approx((0, 0, 1))


def test_rotate_mat3_double():

    A = np.eye(3, dtype=np.float64)
    v1 = np.array([0, 0, 1.0], dtype=np.float64).reshape(3, 1)
    v2 = np.array([1, 1, 1.0], dtype=np.float64).reshape(3, 1)
    R = compute_change_of_basis_operation(v1, v2)
    A = np.matmul(np.matmul(R.T, A), R).reshape(3, 3, 1)

    R = R.T

    rotated = rotate_mat3_double(R, A)

    assert rotated[:, :, 0].flatten() == pytest.approx(
        (1, 0, 0, 0, 1, 0, 0, 0, 1), abs=1e-12
    )


def test_ReflectionLikelihood(testdata):
    def check(
        mosaicity_parameterisation,
        wavelength_parameterisation,
        fix_mosaic_spread=False,
        fix_wavelength_spread=False,
        fix_unit_cell=False,
        fix_orientation=False,
    ):
        experiment = testdata.experiment
        models = testdata.models
        s0 = np.array(testdata.s0)
        sp = np.array(testdata.sp)
        h = testdata.h
        ctot = testdata.ctot
        mobs = testdata.mobs
        Sobs = testdata.Sobs

        U_params = models[1].get_param_vals()
        B_params = models[2].get_param_vals()
        M_params = np.array(models[0][: mosaicity_parameterisation.num_parameters()])
        L_params = flex.double(models[3])

        state = ModelState(
            experiment,
            mosaicity_parameterisation,
            wavelength_parameterisation,
            fix_mosaic_spread=fix_mosaic_spread,
            fix_wavelength_spread=fix_wavelength_spread,
            fix_unit_cell=fix_unit_cell,
            fix_orientation=fix_orientation,
        )
        state.U_params = U_params
        state.B_params = B_params
        state.M_params = M_params
        state.L_params = L_params

        def get_reflection_likelihood(state):
            return ReflectionLikelihood(state, s0, sp, h, ctot, mobs, Sobs)

        likelihood = get_reflection_likelihood(state)

        step = 1e-6

        dL_dp = likelihood.first_derivatives()

        parameters = state.active_parameters

        assert len(dL_dp) == len(parameters)

        def compute_likelihood(parameters):
            state.active_parameters = parameters
            likelihood = get_reflection_likelihood(state)
            return likelihood.log_likelihood()

        dL_num = []
        for i in range(len(parameters)):

            def f(x):
                p = copy.copy(parameters)
                p[i] = x
                return compute_likelihood(p)

            dL_num.append(first_derivative(f, parameters[i], step))

        assert len(dL_num) == len(parameters)
        print(dL_num)
        print(list(dL_dp))
        for n, c in zip(dL_num, dL_dp):
            print(n, c)
            assert n == pytest.approx(c, abs=1e-4)

    S1 = Simple1MosaicityParameterisation()
    S6 = Simple6MosaicityParameterisation()

    check(S1, None, fix_wavelength_spread=True)
    check(S1, None, fix_wavelength_spread=True, fix_mosaic_spread=True)
    check(S1, None, fix_wavelength_spread=True, fix_unit_cell=True)
    check(S1, None, fix_wavelength_spread=True, fix_orientation=True)

    check(S6, None, fix_wavelength_spread=True, fix_mosaic_spread=True)
    check(S6, None, fix_wavelength_spread=True, fix_unit_cell=True)
    check(S6, None, fix_wavelength_spread=True, fix_orientation=True)
    check(S6, None, fix_wavelength_spread=True)


def test_Refiner(testdata, refinerdata_testdata):

    experiment = testdata.experiment
    data = refinerdata_testdata

    sigma_d = 0.02**2
    S1 = Simple1ProfileModel.from_sigma_d(sigma_d).parameterisation()
    S6 = Simple6ProfileModel.from_sigma_d(sigma_d).parameterisation()

    def check(
        parameterisation,
        fix_mosaic_spread=True,
        fix_orientation=True,
        fix_unit_cell=True,
        fix_wavelength_spread=True,
    ):

        state = ModelState(
            experiment,
            parameterisation,
            fix_mosaic_spread=fix_mosaic_spread,
            fix_orientation=fix_orientation,
            fix_unit_cell=fix_unit_cell,
            fix_wavelength_spread=fix_wavelength_spread,
        )

        refiner = Refiner(state, data)
        refiner.refine()

        print(state.mosaicity_covariance_matrix)

    check(S1, fix_mosaic_spread=False, fix_orientation=True, fix_unit_cell=True)
    check(S1, fix_mosaic_spread=True, fix_orientation=False, fix_unit_cell=True)
    check(S1, fix_mosaic_spread=True, fix_orientation=True, fix_unit_cell=False)
    check(S1, fix_mosaic_spread=False, fix_orientation=False, fix_unit_cell=True)
    check(S1, fix_mosaic_spread=True, fix_orientation=False, fix_unit_cell=False)
    check(S1, fix_mosaic_spread=False, fix_orientation=False, fix_unit_cell=False)

    check(S6, fix_mosaic_spread=False, fix_orientation=True, fix_unit_cell=True)
    check(S6, fix_mosaic_spread=True, fix_orientation=False, fix_unit_cell=True)
    check(S6, fix_mosaic_spread=True, fix_orientation=True, fix_unit_cell=False)
    check(S6, fix_mosaic_spread=False, fix_orientation=False, fix_unit_cell=True)
    check(S6, fix_mosaic_spread=True, fix_orientation=False, fix_unit_cell=False)
    check(S6, fix_mosaic_spread=False, fix_orientation=False, fix_unit_cell=False)


def test_RefinerData(testdata):

    experiment = testdata.experiment
    reflections = testdata.reflections

    panel = experiment.detector[0]
    s0_length = matrix.col(experiment.beam.get_s0()).length()
    reflections["bbox"] = flex.int6(len(reflections))
    reflections["xyzobs.px.value"] = flex.vec3_double(len(reflections))
    reflections["s2"] = reflections["s1"].each_normalize() * s0_length
    reflections["sp"] = flex.vec3_double(len(reflections))
    for i, (x, y, z) in enumerate(reflections["xyzcal.px"]):
        x0 = int(x) - 5
        x1 = int(x) + 5 + 1
        y0 = int(y) - 5
        y1 = int(y) + 5 + 1
        z0 = int(z)
        z1 = z0 + 1
        reflections["bbox"][i] = x0, x1, y0, y1, z0, z1
        reflections["xyzobs.px.value"][i] = (int(x) + 0.5, int(y) + 0.5, int(z) + 0.5)
        reflections["sp"][i] = (
            matrix.col(
                panel.get_pixel_lab_coord(reflections["xyzobs.px.value"][i][0:2])
            ).normalize()
            * s0_length
        )

    reflections["shoebox"] = flex.shoebox(
        reflections["panel"], reflections["bbox"], allocate=True
    )

    shoebox_data = flex.float(flex.grid(1, 11, 11))
    shoebox_mask = flex.int(flex.grid(1, 11, 11))
    for j in range(11):
        for i in range(11):
            shoebox_data[0, j, i] = (
                100
                * exp(-0.5 * (j - 5) ** 2 / 1**2)
                * exp(-0.5 * (i - 5) ** 2 / 1**2)
            )
            shoebox_mask[0, j, i] = 5
    for sbox in reflections["shoebox"]:
        sbox.data = shoebox_data
        sbox.mask = shoebox_mask

    data = RefinerData.from_reflections(experiment, reflections)

    assert tuple(data.s0) == pytest.approx(experiment.beam.get_s0())
    assert data.h_list == reflections["miller_index"]
    for i, sp in enumerate(reflections["sp"]):
        assert data.sp_list[:, i] == pytest.approx(sp)
    assert data.ctot_list[0] == sum(shoebox_data)

    mobs1 = np.abs(data.mobs_list[0, :])
    mobs2 = np.abs(data.mobs_list[1, :])
    assert np.max(mobs1) < 1e-6
    assert np.max(mobs2) < 1e-6



from __future__ import annotations

import numpy as np

from dials.array_family import flex


def make_dataset(handle, name, dtype, data, description, units=None):
    dset = handle.create_dataset(
        name, data.focus(), dtype=dtype, data=data.as_numpy_array().astype(dtype)
    )
    dset.attrs["description"] = description
    if units is not None:
        dset.attrs["units"] = units
    return dset


def make_uint(handle, name, data, description, units=None):
    return make_dataset(handle, name, "uint64", data, description, units)


def make_int(handle, name, data, description, units=None):
    return make_dataset(handle, name, "int64", data, description, units)


def make_bool(handle, name, data, description, units=None):
    return make_dataset(handle, name, "int8", data, description, units)


def make_float(handle, name, data, description, units=None):
    return make_dataset(handle, name, "float64", data, description, units)


def make_vlen_uint(handle, name, data, description, units=None):
    import h5py

    dtype = h5py.special_dtype(vlen=np.dtype("uint64"))
    dset = handle.create_dataset(name, (len(data),), dtype=dtype)
    for i, d in enumerate(data):
        if len(d) > 0:
            dset[i] = d
    dset.attrs["description"] = description
    if units is not None:
        dset.attrs["units"] = units
    return dset


def write(handle, key, data):
    if key == "miller_index":
        col1, col2, col3 = zip(*list(data))
        col1 = flex.int(col1)
        col2 = flex.int(col2)
        col3 = flex.int(col3)
        dsc1 = "The h component of the miller index"
        dsc2 = "The k component of the miller index"
        dsc3 = "The l component of the miller index"
        make_int(handle, "h", col1, dsc1)
        make_int(handle, "k", col2, dsc2)
        make_int(handle, "l", col3, dsc3)
    elif key == "id":
        col = data
        dsc = "The experiment id"
        make_int(handle, "id", col, dsc)
    elif key == "partial_id":
        col = data
        desc = "The reflection id"
        make_uint(handle, "reflection_id", col, desc)
    elif key == "entering":
        col = data
        dsc = "Entering or exiting the Ewald sphere"
        make_bool(handle, "entering", col, dsc)
    elif key == "flags":
        col = data
        dsc = "Status of the reflection in processing"
        make_uint(handle, "flags", col, dsc)
    elif key == "panel":
        col = data
        dsc = "The detector module on which the reflection was recorded"
        make_uint(handle, "det_module", col, dsc)
    elif key == "d":
        col = data
        dsc = "The resolution of the reflection"
        make_float(handle, "d", col, dsc)
    elif key == "partiality":
        col = data
        dsc = "The partiality of the reflection"
        make_float(handle, "partiality", col, dsc)
    elif key == "xyzcal.px":
        col1, col2, col3 = data.parts()
        dsc1 = "The predicted bragg peak fast pixel location"
        dsc2 = "The predicted bragg peak slow pixel location"
        dsc3 = "The predicted bragg peak frame number"
        make_float(handle, "predicted_px_x", col1, dsc1)
        make_float(handle, "predicted_px_y", col2, dsc2)
        make_float(handle, "predicted_frame", col3, dsc3)
        handle["predicted_px_x"].attrs["units"] = ""
        handle["predicted_px_y"].attrs["units"] = ""
        handle["predicted_frame"].attrs["units"] = ""
    elif key == "xyzcal.mm":
        col1, col2, col3 = data.parts()
        dsc1 = "The predicted bragg peak fast millimeter location"
        dsc2 = "The predicted bragg peak slow millimeter location"
        dsc3 = "The predicted bragg peak rotation angle number"
        make_float(handle, "predicted_x", col1, dsc1, units="mm")
        make_float(handle, "predicted_y", col2, dsc2, units="mm")
        make_float(handle, "predicted_phi", col3, dsc3, units="rad")
    elif key == "bbox":
        d = data.as_int()
        d.reshape(flex.grid((len(data)), 6))
        make_int(handle, "bounding_box", d, "The reflection bounding box")
        handle["bounding_box"].attrs["units"] = ""
    elif key == "xyzobs.px.value":
        col1, col2, col3 = data.parts()
        dsc1 = "The observed centroid fast pixel value"
        dsc2 = "The observed centroid slow pixel value"
        dsc3 = "The observed centroid frame value"
        make_float(handle, "observed_px_x", col1, dsc1)
        make_float(handle, "observed_px_y", col2, dsc2)
        make_float(handle, "observed_frame", col3, dsc3)
        handle["observed_px_x"].attrs["units"] = ""
        handle["observed_px_y"].attrs["units"] = ""
        handle["observed_frame"].attrs["units"] = ""
    elif key == "xyzobs.px.variance":
        col1, col2, col3 = data.parts()
        dsc1 = "The observed centroid fast pixel variance"
        dsc2 = "The observed centroid slow pixel variance"
        dsc3 = "The observed centroid frame variance"
        make_float(handle, "observed_px_x_var", col1, dsc1)
        make_float(handle, "observed_px_y_var", col2, dsc2)
        make_float(handle, "observed_frame_var", col3, dsc3)
        handle["observed_px_x_var"].attrs["units"] = ""
        handle["observed_px_y_var"].attrs["units"] = ""
        handle["observed_frame_var"].attrs["units"] = ""
    elif key == "xyzobs.mm.value":
        col1, col2, col3 = data.parts()
        dsc1 = "The observed centroid fast pixel value"
        dsc2 = "The observed centroid slow pixel value"
        dsc3 = "The observed centroid phi value"
        make_float(handle, "observed_x", col1, dsc1, units="mm")
        make_float(handle, "observed_y", col2, dsc2, units="mm")
        make_float(handle, "observed_phi", col3, dsc3, units="rad")
    elif key == "xyzobs.mm.variance":
        col1, col2, col3 = data.parts()
        dsc1 = "The observed centroid fast pixel variance"
        dsc2 = "The observed centroid slow pixel variance"
        dsc3 = "The observed centroid phi variance"
        make_float(handle, "observed_x_var", col1, dsc1, units="mm")
        make_float(handle, "observed_y_var", col2, dsc2, units="mm")
        make_float(handle, "observed_phi_var", col3, dsc3, units="rad")
    elif key == "background.mean":
        col = data
        dsc = "The mean background value"
        make_float(handle, "background_mean", col, dsc)
    elif key == "intensity.sum.value":
        col = data
        dsc = "The value of the summed intensity"
        make_float(handle, "int_sum", col, dsc)
    elif key == "intensity.sum.variance":
        col = data
        dsc = "The variance of the summed intensity"
        make_float(handle, "int_sum_var", col, dsc)
    elif key == "intensity.prf.value":
        col = data
        dsc = "The value of the profile fitted intensity"
        make_float(handle, "int_prf", col, dsc)
    elif key == "intensity.prf.variance":
        col = data
        dsc = "The variance of the profile fitted intensity"
        make_float(handle, "int_prf_var", col, dsc)
    elif key == "profile.correlation":
        col = data
        dsc = "Profile fitting correlations"
        make_float(handle, "prf_cc", col, dsc)
    elif key == "lp":
        col = data
        dsc = "The lorentz-polarization correction factor"
        make_float(handle, "lp", col, dsc)
    elif key == "num_pixels.background":
        col = data
        dsc = "Number of background pixels"
        make_int(handle, "num_bg", col, dsc)
    elif key == "num_pixels.background_used":
        col = data
        dsc = "Number of background pixels used"
        make_int(handle, "num_bg_used", col, dsc)
    elif key == "num_pixels.foreground":
        col = data
        dsc = "Number of foreground pixels"
        make_int(handle, "num_fg", col, dsc)
    elif key == "num_pixels.valid":
        col = data
        dsc = "Number of valid pixels"
        make_int(handle, "num_valid", col, dsc)
    elif key == "profile.rmsd":
        col = data
        dsc = "Profile rmsd"
        make_float(handle, "prf_rmsd", col, dsc)
    else:
        raise KeyError(f"Column {key} not written to file")


def read(handle, key):
    from dxtbx.format.nexus import convert_units

    if key == "miller_index":
        h = flex.int(handle["h"][:].astype(np.int32))
        k = flex.int(handle["k"][:].astype(np.int32))
        l = flex.int(handle["l"][:].astype(np.int32))
        return flex.miller_index(h, k, l)
    elif key == "id":
        return flex.int(handle["id"][:].astype(int))
    elif key == "partial_id":
        return flex.size_t(handle["reflection_id"][:].astype(int))
    elif key == "entering":
        return flex.bool(handle["entering"][:].astype(bool))
    elif key == "flags":
        return flex.size_t(handle["flags"][:].astype(int))
    elif key == "panel":
        return flex.size_t(handle["det_module"][:].astype(int))
    elif key == "d":
        return flex.double(handle["d"][:])
    elif key == "partiality":
        return flex.double(handle["partiality"][:])
    elif key == "xyzcal.px":
        x = flex.double(handle["predicted_px_x"][:])
        y = flex.double(handle["predicted_px_y"][:])
        z = flex.double(handle["predicted_frame"][:])
        return flex.vec3_double(x, y, z)
    elif key == "xyzcal.mm":
        x = convert_units(
            flex.double(handle["predicted_x"][:]),
            handle["predicted_x"].attrs["units"],
            "mm",
        )
        y = convert_units(
            flex.double(handle["predicted_y"][:]),
            handle["predicted_y"].attrs["units"],
            "mm",
        )
        z = convert_units(
            flex.double(handle["predicted_phi"][:]),
            handle["predicted_phi"].attrs["units"],
            "rad",
        )
        return flex.vec3_double(x, y, z)
    elif key == "bbox":
        b = flex.int(handle["bounding_box"][:].astype(np.int32))
        return flex.int6(b.as_1d())
    elif key == "xyzobs.px.value":
        x = flex.double(handle["observed_px_x"][:])
        y = flex.double(handle["observed_px_y"][:])
        z = flex.double(handle["observed_frame"][:])
        return flex.vec3_double(x, y, z)
    elif key == "xyzobs.px.variance":
        x = flex.double(handle["observed_px_x_var"][:])
        y = flex.double(handle["observed_px_y_var"][:])
        z = flex.double(handle["observed_frame_var"][:])
        return flex.vec3_double(x, y, z)
    elif key == "xyzobs.mm.value":
        x = convert_units(
            flex.double(handle["observed_x"][:]),
            handle["observed_x"].attrs["units"],
            "mm",
        )
        y = convert_units(
            flex.double(handle["observed_y"][:]),
            handle["observed_y"].attrs["units"],
            "mm",
        )
        z = convert_units(
            flex.double(handle["observed_phi"][:]),
            handle["observed_phi"].attrs["units"],
            "rad",
        )
        return flex.vec3_double(x, y, z)
    elif key == "xyzobs.mm.variance":
        x = convert_units(
            flex.double(handle["observed_x_var"][:]),
            handle["observed_x_var"].attrs["units"],
            "mm",
        )
        y = convert_units(
            flex.double(handle["observed_y_var"][:]),
            handle["observed_y_var"].attrs["units"],
            "mm",
        )
        z = convert_units(
            flex.double(handle["observed_phi_var"][:]),
            handle["observed_phi_var"].attrs["units"],
            "rad",
        )
        return flex.vec3_double(x, y, z)
    elif key == "background.mean":
        return flex.double(handle["background_mean"][:])
    elif key == "intensity.sum.value":
        return flex.double(handle["int_sum"][:])
    elif key == "intensity.sum.variance":
        return flex.double(handle["int_sum_var"][:])
    elif key == "intensity.prf.value":
        return flex.double(handle["int_prf"][:])
    elif key == "intensity.prf.variance":
        return flex.double(handle["int_prf_var"][:])
    elif key == "profile.correlation":
        return flex.double(handle["prf_cc"][:])
    elif key == "lp":
        return flex.double(handle["lp"][:])
    elif key == "num_pixels.background":
        return flex.int(handle["num_bg"][:].astype(np.int32))
    elif key == "num_pixels.background_used":
        return flex.int(handle["num_bg_used"][:].astype(np.int32))
    elif key == "num_pixels.foreground":
        return flex.int(handle["num_fg"][:].astype(np.int32))
    elif key == "num_pixels.valid":
        return flex.int(handle["num_valid"][:].astype(np.int32))
    elif key == "profile.rmsd":
        return flex.double(handle["prf_rmsd"][:])
    else:
        raise KeyError(f"Column {key} not read from file")


def dump(entry, reflections, experiments):
    print("Dumping NXreflections")

    # Add the feature
    if "features" in entry:
        features = entry["features"]
        assert features.dtype == "uint64"
        features.resize((len(features) + 1,))
        features[len(features) - 1] = 7
    else:
        features = entry.create_dataset(
            "features", (1,), maxshape=(None,), dtype=np.uint64
        )
        features[0] = 7

    # Create the base class
    assert "reflections" not in entry
    refls = entry.create_group("reflections")
    refls.attrs["NX_class"] = "NXreflections"
    refls.attrs["description"] = ""

    refls["experiments"] = [np.string_(e) for e in experiments]

    if reflections is None:
        return

    # For each column in the reflection table dump to file
    for key, data in reflections.cols():
        try:
            write(refls, key, data)
        except KeyError as e:
            print(e.args[0])

    # FIXME Write the overlaps (for testing at the moment)
    # Optional and doesn't pass validation, so disable
    # overlaps = [[] for i in range(len(reflections))]
    # overlaps[0] = [1, 2, 3]
    # overlaps[1] = [0, 4]
    # overlaps[2] = [0, 3]
    # overlaps[3] = [0, 2]
    # overlaps[4] = [1]
    # make_vlen_uint(refls, "overlaps", overlaps, "Reflection overlap list")


def load(entry):
    print("Loading NXreflections")

    # Check the feature is present
    assert "features" in entry
    assert 7 in entry["features"]

    # Get the entry
    refls = entry["reflections"]
    if refls.attrs["NX_class"] == "NXsubentry":
        # Backward compatibility. See https://github.com/nexusformat/definitions/pull/752
        # Get the definition
        definition = refls["definition"]
        assert definition[()] == "NXreflections"
        assert definition.attrs["version"] == 1
    else:
        assert refls.attrs["NX_class"] == "NXreflections"

    # The paths to the experiments
    experiments = list(refls["experiments"])

    # The columns to try
    columns = [
        "miller_index",
        "id",
        "partial_id",
        "entering",
        "flags",
        "panel",
        "d",
        "partiality",
        "xyzcal.px",
        "xyzcal.mm",
        "bbox",
        "xyzobs.px.value",
        "xyzobs.px.variance",
        "xyzobs.mm.value",
        "xyzobs.mm.variance",
        "background.mean",
        "intensity.sum.value",
        "intensity.sum.variance",
        "intensity.prf.value",
        "intensity.prf.variance",
        "profile.correlation",
        "lp",
        "num_pixels.background",
        "num_pixels.foreground",
        "num_pixels.background_used",
        "num_pixels.valid",
        "profile.rmsd",
    ]

    # The reflection table
    table = None

    # For each column in the reflection table dump to file
    for key in columns:
        try:
            col = read(refls, key)
        except KeyError:
            continue
        if table is None:
            table = flex.reflection_table()
        table[key] = col

    # Return the table
    return table, experiments



from __future__ import annotations

from dials.algorithms.background.simple.algorithm import BackgroundAlgorithm
from dials_algorithms_background_simple_ext import *  # noqa: F403; lgtm

__all__ = (  # noqa: F405
    "BackgroundAlgorithm",
    "Constant2dModel",
    "Constant2dModeller",
    "Constant3dModel",
    "Constant3dModeller",
    "Creator",
    "Linear2dModel",
    "Linear2dModeller",
    "Linear3dModel",
    "Linear3dModeller",
    "Model",
    "Modeller",
    "MosflmOutlierRejector",
    "NSigmaOutlierRejector",
    "NormalOutlierRejector",
    "OutlierRejector",
    "TruncatedOutlierRejector",
    "TukeyOutlierRejector",
    "absolute_maximum_n_sigma",
    "is_normally_distributed",
    "maximum_n_sigma",
    "minimum_n_sigma",
    "normal_expected_n_sigma",
)



from __future__ import annotations

import json

import procrunner
import pytest

import iotbx.cif
from cctbx import sgtbx, uctbx
from dxtbx.model.experiment_list import ExperimentListFactory
from dxtbx.serialize import load
from iotbx import mtz

from dials.array_family import flex
from dials.command_line.slice_sequence import slice_experiments, slice_reflections
from dials.util.multi_dataset_handling import assign_unique_identifiers


def run_export(export_format, dials_data, tmp_path):
    result = procrunner.run(
        [
            "dials.export",
            "format=" + export_format,
            dials_data("centroid_test_data", pathlib=True) / "experiments.json",
            dials_data("centroid_test_data", pathlib=True) / "integrated.pickle",
        ],
        working_directory=tmp_path,
    )
    assert not result.returncode and not result.stderr
    assert (tmp_path / f"integrated.{export_format}").is_file()


def test_nxs(dials_data, tmp_path):
    run_export("nxs", dials_data, tmp_path)


def test_mtz(dials_data, tmp_path):
    result = procrunner.run(
        [
            "dials.export",
            "format=mtz",
            "project_name=ham",
            "crystal_name=spam",
            dials_data("centroid_test_data", pathlib=True) / "experiments.json",
            dials_data("centroid_test_data", pathlib=True) / "integrated.pickle",
        ],
        working_directory=tmp_path,
    )
    assert not result.returncode and not result.stderr
    assert (tmp_path / "integrated.mtz").is_file()
    mtz_obj = mtz.object(str(tmp_path / "integrated.mtz"))
    assert mtz_obj.crystals()[1].name() == "spam"
    assert mtz_obj.crystals()[1].project_name() == "ham"


def test_mtz_recalculated_cell(dials_data, tmp_path):
    # First run dials.two_theta_refine to ensure that the crystals have
    # recalculated_unit_cell set
    scaled_expt = (
        dials_data("x4wide_processed", pathlib=True) / "AUTOMATIC_DEFAULT_scaled.expt"
    )
    scaled_refl = (
        dials_data("x4wide_processed", pathlib=True) / "AUTOMATIC_DEFAULT_scaled.refl"
    )
    result = procrunner.run(
        ["dials.two_theta_refine", scaled_expt, scaled_refl],
        working_directory=tmp_path,
    )
    assert (tmp_path / "refined_cell.expt").is_file()
    refined_expt = load.experiment_list(
        tmp_path / "refined_cell.expt", check_format=False
    )
    ttr_cell = refined_expt.crystals()[0].get_recalculated_unit_cell()

    d_min = 1.3
    result = procrunner.run(
        [
            "dials.export",
            "format=mtz",
            tmp_path / "refined_cell.expt",
            scaled_refl,
            f"d_min={d_min:f}",
        ],
        working_directory=tmp_path,
    )
    assert not result.returncode and not result.stderr
    assert (tmp_path / "scaled.mtz").is_file()
    # The resulting mtz should have the same unit cell set as the recalculated_unit_cell
    # from dials.two_theta_refine
    for ma in mtz.object(str(tmp_path / "scaled.mtz")).as_miller_arrays():
        assert ttr_cell.parameters() == pytest.approx(ma.unit_cell().parameters())
        assert ma.d_min() >= d_min


def test_mtz_best_unit_cell(dials_data, tmp_path):
    scaled_expt = (
        dials_data("x4wide_processed", pathlib=True) / "AUTOMATIC_DEFAULT_scaled.expt"
    )
    scaled_refl = (
        dials_data("x4wide_processed", pathlib=True) / "AUTOMATIC_DEFAULT_scaled.refl"
    )
    best_unit_cell = uctbx.unit_cell((42, 42, 39, 90, 90, 90))
    d_min = 1.5
    result = procrunner.run(
        [
            "dials.export",
            "format=mtz",
            scaled_expt,
            scaled_refl,
            f"d_min={d_min:f}",
            "best_unit_cell=%g,%g,%g,%g,%g,%g" % best_unit_cell.parameters(),
        ],
        working_directory=tmp_path,
    )
    assert not result.returncode and not result.stderr
    assert (tmp_path / "scaled.mtz").is_file()
    # The resulting mtz should have the best_unit_cell as input to dials.export
    for ma in mtz.object(str(tmp_path / "scaled.mtz")).as_miller_arrays():
        assert best_unit_cell.parameters() == pytest.approx(ma.unit_cell().parameters())
        assert ma.d_min() >= d_min


def test_multi_sequence_integrated_mtz(dials_data, tmp_path):
    """Test dials.export on multi-sequence integrated data."""
    # first combine two integrated files
    data = dials_data("multi_crystal_proteinase_k", pathlib=True)
    result = procrunner.run(
        [
            "dials.combine_experiments",
            data / "experiments_1.json",
            data / "reflections_1.pickle",
            data / "experiments_2.json",
            data / "reflections_2.pickle",
        ],
        working_directory=tmp_path,
    )

    assert not result.returncode and not result.stderr
    assert (tmp_path / "combined.refl").is_file()
    assert (tmp_path / "combined.expt").is_file()

    # now export
    result = procrunner.run(
        [
            "dials.export",
            "format=mtz",
            "mtz.hklout=integrated.mtz",
            tmp_path / "combined.refl",
            tmp_path / "combined.expt",
        ],
        working_directory=tmp_path,
    )
    assert not result.returncode and not result.stderr
    assert (tmp_path / "integrated.mtz").is_file()


def test_mtz_multi_wavelength(dials_data, tmp_path):
    """Test multi-wavelength mtz export"""
    # First make suitable input - multi datasets experiment list and reflection
    # table with different wavelengths
    mcp = dials_data("multi_crystal_proteinase_k", pathlib=True)
    exp_1 = load.experiment_list(mcp / "experiments_1.json", check_format=False)
    exp_2 = load.experiment_list(mcp / "experiments_2.json", check_format=False)
    refl_1 = flex.reflection_table.from_file(mcp / "reflections_1.pickle")
    refl_2 = flex.reflection_table.from_file(mcp / "reflections_2.pickle")

    exp_1[0].beam.set_wavelength(0.5)
    exp_2[0].beam.set_wavelength(1.0)

    exp_1.extend(exp_2)
    reflection_list = [refl_1, refl_2]
    exps, refls = assign_unique_identifiers(exp_1, reflection_list)
    joint_refl = flex.reflection_table()
    for r in refls:
        joint_refl.extend(r)
    exps.as_json(tmp_path / "tmp_exp.expt")
    joint_refl.as_file(tmp_path / "tmp_refl.refl")

    # Now run
    result = procrunner.run(
        [
            "dials.export",
            tmp_path / "tmp_exp.expt",
            tmp_path / "tmp_refl.refl",
            "format=mtz",
            "mtz.hklout=unmerged.mtz",
        ],
        environment_override={"DIALS_EXPORT_DO_NOT_CHECK_FORMAT": "True"},
        working_directory=tmp_path,
    )
    assert not result.returncode and not result.stderr
    assert (tmp_path / "unmerged.mtz").is_file()

    # Inspect output
    m = mtz.object(str(tmp_path / "unmerged.mtz")).crystals()
    n_batches = []
    wavelengths = []
    for crystal in m:
        for dataset in crystal.datasets():
            wavelengths.append(dataset.wavelength())
            n_batches.append(dataset.n_batches())
    assert n_batches == [0, 25, 25]  # base, dataset1, dataset2
    assert wavelengths == [0, 0.5, 1.0]  # base, dataset1, dataset2


def test_mtz_primitive_cell(dials_data, tmp_path):
    scaled_expt = dials_data("insulin_processed", pathlib=True) / "scaled.expt"
    scaled_refl = dials_data("insulin_processed", pathlib=True) / "scaled.refl"

    # First reindex to the primitive setting
    expts = load.experiment_list(scaled_expt, check_format=False)
    cs = expts[0].crystal.get_crystal_symmetry()
    cb_op = cs.change_of_basis_op_to_primitive_setting()
    procrunner.run(
        [
            "dials.reindex",
            scaled_expt,
            scaled_refl,
            f'change_of_basis_op="{cb_op}"',
        ],
        working_directory=tmp_path,
    )

    # Now export the reindexed experiments/reflections
    procrunner.run(
        ["dials.export", tmp_path / "reindexed.expt", tmp_path / "reindexed.refl"],
        working_directory=tmp_path,
    )

    mtz_obj = mtz.object(str(tmp_path / "scaled.mtz"))
    cs_primitive = cs.change_basis(cb_op)
    assert mtz_obj.space_group() == cs_primitive.space_group()
    refl = flex.reflection_table.from_file(scaled_refl)
    refl = refl.select(~refl.get_flags(refl.flags.bad_for_scaling, all=False))
    for ma in mtz_obj.as_miller_arrays():
        assert ma.crystal_symmetry().is_similar_symmetry(cs_primitive)
        assert ma.d_max_min() == pytest.approx(
            (flex.max(refl["d"]), flex.min(refl["d"]))
        )


@pytest.mark.parametrize("compress", [None, "gz", "bz2", "xz"])
@pytest.mark.parametrize("hklout", [None, "my.cif"])
def test_mmcif(compress, hklout, dials_data, tmp_path):
    # Call dials.export after integration

    command = [
        "dials.export",
        "format=mmcif",
        dials_data("centroid_test_data", pathlib=True) / "experiments.json",
        dials_data("centroid_test_data", pathlib=True) / "integrated.pickle",
    ]
    if hklout is not None:
        command.append(f"mmcif.hklout={hklout}")
    else:
        hklout = "integrated.cif"
    if compress is not None:
        command.append(f"mmcif.compress={compress}")
        hklin = hklout + "." + compress
    else:
        hklin = hklout
    result = procrunner.run(command, working_directory=tmp_path)
    assert not result.returncode and not result.stderr
    assert (tmp_path / hklin).is_file()


@pytest.mark.parametrize("pdb_version", ["v5", "v5_next"])
def test_mmcif_on_scaled_data(dials_data, tmp_path, pdb_version):
    """Call dials.export format=mmcif after scaling"""
    scaled_expt = (
        dials_data("x4wide_processed", pathlib=True) / "AUTOMATIC_DEFAULT_scaled.expt"
    )
    scaled_refl = (
        dials_data("x4wide_processed", pathlib=True) / "AUTOMATIC_DEFAULT_scaled.refl"
    )
    command = [
        "dials.export",
        "format=mmcif",
        scaled_expt,
        scaled_refl,
        "mmcif.hklout=scaled.mmcif",
        "compress=None",
        f"pdb_version={pdb_version}",
    ]
    result = procrunner.run(command, working_directory=tmp_path)
    assert not result.returncode and not result.stderr
    assert (tmp_path / "scaled.mmcif").is_file()

    model = iotbx.cif.reader(file_path=str(tmp_path / "scaled.mmcif")).model()
    if pdb_version == "v5":
        assert "_pdbx_diffrn_data_section.id" not in model["dials"].keys()
    elif pdb_version == "v5_next":
        assert "_pdbx_diffrn_data_section.id" in model["dials"].keys()


def test_mmcif_p1_narrow_wedge(dials_data, tmp_path):
    """Call dials.export format=mmcif after scaling"""
    data_dir = dials_data("x4wide_processed", pathlib=True)

    refl = flex.reflection_table.from_file(data_dir / "AUTOMATIC_DEFAULT_scaled.refl")
    refl = slice_reflections(refl, [(1, 3)])
    refl.as_file(tmp_path / "p1_narrow.refl")

    expts = load.experiment_list(
        data_dir / "AUTOMATIC_DEFAULT_scaled.expt", check_format=False
    )
    expts = slice_experiments(expts, [(1, 3)])
    expts[0].crystal.set_space_group(sgtbx.space_group())
    expts.as_file(tmp_path / "p1_narrow.expt")

    command = [
        "dials.export",
        "format=mmcif",
        tmp_path / "p1_narrow.expt",
        tmp_path / "p1_narrow.refl",
        "mmcif.hklout=scaled.mmcif",
        "compress=None",
    ]
    result = procrunner.run(command, working_directory=tmp_path)
    assert not result.returncode and not result.stderr
    assert (tmp_path / "scaled.mmcif").is_file()

    model = iotbx.cif.reader(file_path=str(tmp_path / "scaled.mmcif")).model()
    assert model["dials"]["_reflns.pdbx_redundancy"] == "1.0"
    assert model["dials"]["_reflns.pdbx_CC_half"] == "0.0"


def test_xds_ascii(dials_data, tmp_path):
    # Call dials.export
    result = procrunner.run(
        [
            "dials.export",
            "intensity=sum",
            "format=xds_ascii",
            dials_data("centroid_test_data", pathlib=True) / "experiments.json",
            dials_data("centroid_test_data", pathlib=True) / "integrated.pickle",
        ],
        working_directory=tmp_path,
    )
    assert not result.returncode and not result.stderr
    assert (tmp_path / "DIALS.HKL").is_file()

    psi_values = {
        (-9, 7, -10): 153.430361,
        (-5, 11, -26): 175.559441,
        (-4, 23, 24): 129.468070,
        (2, 10, 20): 147.947274,
    }

    with (tmp_path / "DIALS.HKL").open() as fh:
        for record in fh:
            if record.startswith("!"):
                continue
            tokens = record.split()
            hkl = tuple(map(int, tokens[:3]))
            if hkl not in psi_values:
                continue
            psi = float(tokens[-1])
            assert psi == pytest.approx(psi_values[hkl], abs=0.1)


def test_sadabs(dials_data, tmp_path):
    # Call dials.export
    result = procrunner.run(
        [
            "dials.export",
            "intensity=sum",
            "mtz.partiality_threshold=0.99",
            "format=sadabs",
            dials_data("centroid_test_data", pathlib=True) / "experiments.json",
            dials_data("centroid_test_data", pathlib=True) / "integrated.pickle",
        ],
        working_directory=tmp_path,
    )
    assert not result.returncode and not result.stderr
    assert (tmp_path / "integrated.sad").is_file()

    direction_cosines = {
        (-9, 7, -10): (0.51253, -0.72107, 0.84696, -0.68476, -0.14130, -0.10561),
        (-5, 11, -26): (0.51310, -0.62895, 0.84711, -0.59223, -0.13830, -0.50366),
        (-4, 23, 24): (0.51308, -0.60578, 0.84711, -0.31416, -0.13840, 0.73099),
        (2, 10, 20): (0.51239, -0.46605, 0.84693, -0.61521, -0.14204, 0.63586),
    }

    with (tmp_path / "integrated.sad").open() as fh:
        for record in fh:
            record = record.replace("-", " -")
            tokens = record.split()
            hkl = tuple(map(int, tokens[:3]))
            cosines = tuple(map(float, tokens[6:12]))
            if hkl not in direction_cosines:
                continue
            assert cosines == pytest.approx(direction_cosines[hkl], abs=0.001)


def test_json(dials_data, tmp_path):
    # Call dials.export
    result = procrunner.run(
        [
            "dials.export",
            "format=json",
            dials_data("centroid_test_data", pathlib=True)
            / "imported_experiments.json",
            dials_data("centroid_test_data", pathlib=True) / "strong.pickle",
        ],
        working_directory=tmp_path,
    )
    assert not result.returncode and not result.stderr
    assert (tmp_path / "rlp.json").is_file()

    with open(tmp_path / "rlp.json", mode="rb") as fh:
        d = json.load(fh)
    assert set(d) == {"imageset_id", "experiments", "rlp", "experiment_id"}
    assert d["rlp"][:3] == [0.123413, 0.576679, 0.186326], d["rlp"][:3]
    assert d["imageset_id"][0] == 0
    assert d["experiment_id"][0] == 0
    experiments = ExperimentListFactory.from_dict(d["experiments"])
    imgset = experiments.imagesets()
    assert len(imgset) == 1


def test_json_shortened(dials_data, tmp_path):
    # Call dials.export
    result = procrunner.run(
        [
            "dials.export",
            "format=json",
            dials_data("centroid_test_data", pathlib=True) / "experiments.json",
            dials_data("centroid_test_data", pathlib=True) / "integrated.pickle",
            "json.filename=integrated.json",
            "n_digits=4",
            "compact=False",
        ],
        working_directory=tmp_path,
    )
    assert not result.returncode and not result.stderr
    assert (tmp_path / "integrated.json").is_file()

    with open(tmp_path / "integrated.json", mode="rb") as fh:
        d = json.load(fh)
    assert "imageset_id" in d
    assert "rlp" in d
    assert "experiment_id" in d
    assert d["rlp"][:3] == [-0.5975, -0.6141, 0.4702], d["rlp"][:3]
    assert d["imageset_id"][0] == 0
    assert d["experiment_id"][0] == 0


def test_shelx(dials_data, tmp_path):
    # Call dials.export
    result = procrunner.run(
        [
            "dials.export",
            "intensity=scale",
            "format=shelx",
            dials_data("l_cysteine_4_sweeps_scaled", pathlib=True)
            / "scaled_20_25.expt",
            dials_data("l_cysteine_4_sweeps_scaled", pathlib=True)
            / "scaled_20_25.refl",
        ],
        working_directory=tmp_path,
    )
    assert not result.returncode and not result.stderr
    assert (tmp_path / "dials.hkl").is_file()

    intensities_sigmas = {
        (4, 2, 4): (173.14, 15.39),
        (3, -3, -3): (324.13, 25.92),
        (4, 0, 2): (876.02, 69.34),
        (3, -2, -1): (463.11, 36.76),
    }

    with (tmp_path / "dials.hkl").open() as fh:
        max_intensity = -9999.0
        for record in fh:
            tokens = record.split()
            hkl = tuple(map(int, tokens[:3]))
            i_sigi = tuple(map(float, tokens[3:5]))
            if hkl not in intensities_sigmas:
                if i_sigi[0] > max_intensity:
                    max_intensity = i_sigi[0]
                continue
            assert i_sigi == pytest.approx(intensities_sigmas[hkl], abs=0.001)
        assert max_intensity == pytest.approx(9999.00, abs=0.001)


def test_shelx_ins(dials_data, tmp_path):
    # Call dials.export
    result = procrunner.run(
        [
            "dials.export",
            "intensity=scale",
            "format=shelx",
            dials_data("l_cysteine_4_sweeps_scaled", pathlib=True)
            / "scaled_20_25.expt",
            dials_data("l_cysteine_4_sweeps_scaled", pathlib=True)
            / "scaled_20_25.refl",
        ],
        working_directory=tmp_path,
    )
    assert not result.returncode and not result.stderr
    assert (tmp_path / "dials.ins").is_file()

    cell_esds = {
        "CELL": (5.4815, 8.2158, 12.1457, 90.000, 90.000, 90.000),
        "ZERR": (0.0005, 0.0007, 0.0011, 0.003, 0.004, 0.004),
    }

    with (tmp_path / "dials.ins").open() as fh:
        for line in fh:
            tokens = line.split()
            instruction = tokens[0]
            if instruction in cell_esds:
                result = tuple(map(float, tokens[2:8]))
                assert result == pytest.approx(cell_esds[instruction], abs=0.001)


def test_shelx_ins_best_unit_cell(dials_data, tmp_path):
    # Call dials.export
    result = procrunner.run(
        [
            "dials.export",
            "intensity=scale",
            "format=shelx",
            "best_unit_cell=5,8,12,90,90,90",
            dials_data("l_cysteine_4_sweeps_scaled", pathlib=True)
            / "scaled_20_25.expt",
            dials_data("l_cysteine_4_sweeps_scaled", pathlib=True)
            / "scaled_20_25.refl",
        ],
        working_directory=tmp_path,
    )
    assert not result.returncode and not result.stderr
    assert (tmp_path / "dials.ins").is_file()

    cell_esds = {
        "CELL": (5.0, 8.0, 12.0, 90.0, 90.0, 90.0),
    }

    with (tmp_path / "dials.ins").open() as fh:
        for line in fh:
            tokens = line.split()
            instruction = tokens[0]
            assert instruction != "ZERR"
            if instruction in cell_esds:
                result = tuple(map(float, tokens[2:8]))
                assert result == pytest.approx(cell_esds[instruction], abs=0.001)


def test_export_sum_or_profile_only(dials_data, tmp_path):
    expt = dials_data("insulin_processed", pathlib=True) / "integrated.expt"
    refl = dials_data("insulin_processed", pathlib=True) / "integrated.refl"

    for remove in "prf", "sum":
        removed = tmp_path / f"removed_{remove}.refl"
        data = flex.reflection_table.from_file(refl)
        del data[f"intensity.{remove}.value"]
        del data[f"intensity.{remove}.variance"]
        data.as_file(removed)
        result = procrunner.run(
            ["dials.export", expt, removed, f"mtz.hklout=removed_{remove}.mtz"],
            working_directory=tmp_path,
        )
        assert not result.returncode and not result.stderr
        assert (tmp_path / f"removed_{remove}.mtz").is_file()


@pytest.mark.parametrize("intensity_choice", ["profile", "sum"])
def test_pets(dials_data, tmp_path, intensity_choice):
    expt = dials_data("quartz_processed", pathlib=True) / "integrated.expt"
    refl = dials_data("quartz_processed", pathlib=True) / "integrated.refl"
    # Call dials.export
    result = procrunner.run(
        [
            "dials.export",
            "intensity=scale",
            "format=pets",
            "id=0",
            "step=1",
            "n_merged=2",
            "intensity=" + intensity_choice,
            "filename_prefix=" + intensity_choice,
            expt,
            refl,
        ],
        working_directory=tmp_path,
    )
    assert not result.returncode and not result.stderr
    output = tmp_path / (intensity_choice + ".cif_pets")
    # On windows, \r\n endings are written; but our reference is \n
    output_data = output.read_bytes().replace(b"\r\n", b"\n")

    if intensity_choice == "profile":
        reference = dials_data("quartz_processed", pathlib=True) / "dials_prf.cif_pets"
    else:
        reference = dials_data("quartz_processed", pathlib=True) / "dials_dyn.cif_pets"
    assert output_data == reference.read_bytes()



"""Lattice search strategies."""


from __future__ import annotations


class Strategy:
    """A base class for lattice search strategies."""

    phil_help = None

    phil_scope = None

    def __init__(self, params=None, *args, **kwargs):
        """Construct the strategy.

        Args:
            params: an extracted PHIL scope containing the parameters
        """
        self._params = params
        if self._params is None and self.phil_scope is not None:
            self._params = self.phil_scope.extract()

    def find_crystal_models(self, reflections, experiments):
        """Find a list of likely crystal models.

        Args:
            reflections (dials.array_family.flex.reflection_table):
                The found spots centroids and associated data

            experiments (dxtbx.model.experiment_list.ExperimentList):
                The experimental geometry models

        Returns:
            A list of candidate crystal models.
        """
        raise NotImplementedError()



from __future__ import annotations

import logging
import math

from cctbx.sgtbx.bravais_types import bravais_lattice
from dxtbx.model import Crystal
from scitbx.array_family import flex

from dials.algorithms.indexing.compare_orientation_matrices import (
    difference_rotation_matrix_axis_angle,
)
from dials.algorithms.indexing.symmetry import find_matching_symmetry

logger = logging.getLogger(__name__)


def candidate_orientation_matrices(basis_vectors, max_combinations=None):

    # select unique combinations of input vectors to test
    # the order of combinations is such that combinations comprising vectors
    # nearer the beginning of the input list will appear before combinations
    # comprising vectors towards the end of the list
    n = len(basis_vectors)
    # hardcoded limit on number of vectors, fixes issue #72
    # https://github.com/dials/dials/issues/72
    n = min(n, 100)
    basis_vectors = basis_vectors[:n]
    combinations = flex.vec3_int(flex.nested_loop((n, n, n)))
    combinations = combinations.select(
        flex.sort_permutation(combinations.as_vec3_double().norms())
    )

    # select only those combinations where j > i and k > j
    i, j, k = combinations.as_vec3_double().parts()
    sel = flex.bool(len(combinations), True)
    sel &= j > i
    sel &= k > j
    combinations = combinations.select(sel)

    if max_combinations is not None and max_combinations < len(combinations):
        combinations = combinations[:max_combinations]

    half_pi = 0.5 * math.pi
    min_angle = 20 / 180 * math.pi  # 20 degrees, arbitrary cutoff
    for i, j, k in combinations:
        a = basis_vectors[i]
        b = basis_vectors[j]
        angle = a.angle(b)
        if angle < min_angle or (math.pi - angle) < min_angle:
            continue
        a_cross_b = a.cross(b)
        gamma = a.angle(b)
        if gamma < half_pi:
            # all angles obtuse if possible please
            b = -b
            a_cross_b = -a_cross_b
        c = basis_vectors[k]
        if abs(half_pi - a_cross_b.angle(c)) < min_angle:
            continue
        alpha = b.angle(c, deg=True)
        if alpha < half_pi:
            c = -c
        if a_cross_b.dot(c) < 0:
            # we want right-handed basis set, therefore invert all vectors
            a = -a
            b = -b
            c = -c
        model = Crystal(a, b, c, space_group_symbol="P 1")
        uc = model.get_unit_cell()
        cb_op_to_niggli = uc.change_of_basis_op_to_niggli_cell()
        model = model.change_basis(cb_op_to_niggli)

        uc = model.get_unit_cell()
        params = uc.parameters()
        if uc.volume() > (params[0] * params[1] * params[2] / 100):
            # unit cell volume cutoff from labelit 2004 paper
            yield model


def filter_known_symmetry(
    crystal_models,
    target_symmetry,
    relative_length_tolerance=0.1,
    absolute_angle_tolerance=5,
    max_delta=5,
):
    """Filter crystal models for known symmetry.

    Args:
        crystal_models (list): A list of :class:`dxtbx.model.Crystal` objects.
        target_symmetry (cctbx.crystal.symmetry): The target symmetry for filtering.
        relative_length_tolerance (float): Relative tolerance for unit cell lengths in
            unit cell comparison (default value is 0.1).
        absolute_angle_tolerance (float): Angular tolerance (in degrees) in unit cell
            comparison (default value is 5).
        max_delta (float): Maximum allowed Le Page delta used in searching for basis
            vector combinations that are consistent with the given symmetry (default
            value is 5).
    """

    n_matched = 0

    cb_op_ref_to_primitive = target_symmetry.change_of_basis_op_to_primitive_setting()

    if target_symmetry.unit_cell() is not None:
        target_symmetry_primitive = target_symmetry.change_basis(cb_op_ref_to_primitive)
    else:
        target_symmetry_primitive = target_symmetry.customized_copy(
            space_group_info=target_symmetry.space_group_info().change_basis(
                cb_op_ref_to_primitive
            )
        )
    target_bravais_str = str(
        bravais_lattice(
            group=target_symmetry_primitive.space_group_info()
            .reference_setting()
            .group()
        )
    )

    for model in crystal_models:
        uc = model.get_unit_cell()
        best_subgroup = find_matching_symmetry(
            uc, None, max_delta=max_delta, target_bravais_str=target_bravais_str
        )
        if best_subgroup is not None:
            if target_symmetry.unit_cell() is not None and not (
                best_subgroup["best_subsym"]
                .unit_cell()
                .is_similar_to(
                    target_symmetry.as_reference_setting().best_cell().unit_cell(),
                    relative_length_tolerance=relative_length_tolerance,
                    absolute_angle_tolerance=absolute_angle_tolerance,
                )
            ):
                logger.debug(
                    "Rejecting crystal model inconsistent with input symmetry:\n"
                    f"  Unit cell: {str(model.get_unit_cell())}"
                )
                continue

            n_matched += 1
            yield model
    if not n_matched:
        logger.warning(
            "No crystal models remaining after comparing with known symmetry"
        )


def filter_similar_orientations(
    crystal_models, other_crystal_models, minimum_angular_separation=5
):
    for cryst in crystal_models:
        orientation_too_similar = False
        for cryst_a in other_crystal_models:
            R_ab, axis, angle, cb_op_ab = difference_rotation_matrix_axis_angle(
                cryst_a, cryst
            )
            if abs(angle) < minimum_angular_separation:  # degrees
                orientation_too_similar = True
                break
        if orientation_too_similar:
            logger.debug("skipping crystal: too similar to other crystals")
            continue
        yield cryst



"""
Tests for the dials.algorithms.scaling.plots module
"""

from __future__ import annotations

import numpy as np

from dials.algorithms.scaling.model.model import (
    ArrayScalingModel,
    PhysicalScalingModel,
    plot_scaling_models,
)
from dials.algorithms.scaling.plots import (
    normal_probability_plot,
    plot_array_absorption_plot,
    plot_array_decay_plot,
    plot_array_modulation_plot,
    plot_outliers,
)


def test_plot_array_absorption_plot():
    """Test plot generation for array absorption_correction"""
    array_dict = {
        "__id__": "array",
        "is_scaled": True,
        "absorption": {
            "n_parameters": 45,
            "parameters": list(range(1, 46)),
            "est_standard_devs": [],
        },
        "configuration_parameters": {
            "corrections": ["absorption"],
            "time_rot_interval": 10.0,
            "n_x_param": 3,
            "n_y_param": 3,
            "n_time_param": 5,
            "xmin": 0,
            "x_bin_width": 2,
            "ymin": 1,
            "y_bin_width": 2,
            "valid_osc_range": [0, 100],
        },
    }

    array_model = ArrayScalingModel.from_dict(array_dict)
    plot = plot_array_absorption_plot(array_model)
    assert plot["array_absorption_plot"]["data"][0]["x"]
    assert plot["array_absorption_plot"]["data"][0]["y"]
    assert plot["array_absorption_plot"]["data"][0]["z"]


def test_plot_array_decay_plot():
    """Test plot generation for array decay correction"""
    array_dict = {
        "__id__": "array",
        "is_scaled": True,
        "decay": {
            "n_parameters": 20,
            "parameters": list(range(1, 21)),
            "est_standard_devs": [],
        },
        "configuration_parameters": {
            "corrections": ["decay"],
            "time_rot_interval": 10.0,
            "n_res_param": 5,
            "res_bin_width": 0.1,
            "n_time_param": 4,
            "resmin": 0.05,
            "valid_osc_range": [0, 100],
        },
    }

    array_model = ArrayScalingModel.from_dict(array_dict)
    plot = plot_array_decay_plot(array_model)
    assert plot["array_decay_plot"]["data"][0]["x"]
    assert plot["array_decay_plot"]["data"][0]["y"]


def test_plot_array_modulation_plot():
    """Test plot generation for array modulation correction"""
    array_dict = {
        "__id__": "array",
        "is_scaled": True,
        "modulation": {
            "n_parameters": 25,
            "parameters": list(range(1, 26)),
            "est_standard_devs": [],
        },
        "configuration_parameters": {
            "corrections": ["modulation", "decay"],
            "n_x_mod_param": 5,
            "n_y_mod_param": 5,
            "xmin": 0,
            "x_det_bin_width": 2,
            "ymin": 1,
            "y_det_bin_width": 2,
            "time_rot_interval": 10.0,
            "n_res_param": 5,
            "res_bin_width": 0.1,
            "n_time_param": 4,
            "resmin": 0.05,
            "valid_osc_range": [0, 100],
        },
        "decay": {
            "n_parameters": 20,
            "parameters": list(range(1, 21)),
            "est_standard_devs": [],
        },
    }

    array_model = ArrayScalingModel.from_dict(array_dict)
    plot = plot_array_modulation_plot(array_model)
    assert plot["array_modulation_plot"]["data"][0]["x"]
    assert plot["array_modulation_plot"]["data"][0]["y"]


def test_plot_scaling_models():

    physical_dict = {
        "__id__": "physical",
        "is_scaled": True,
        "scale": {
            "n_parameters": 2,
            "parameters": [0.5, 1.0],
            "est_standard_devs": [0.05, 0.1],
        },
        "configuration_parameters": {
            "corrections": ["scale", "decay", "absorption"],
            "s_norm_fac": 0.1,
            "d_norm_fac": 0.1,
            "scale_rot_interval": 10.0,
            "decay_rot_interval": 10.0,
            "decay_restaint": 1e-1,
            "valid_osc_range": [0.0, 2.0],
        },
        "decay": {
            "n_parameters": 2,
            "parameters": [0.5, 1.0],
            "est_standard_devs": [0.05, 0.1],
        },
        "absorption": {
            "n_parameters": 4,
            "parameters": [0.1, -0.1, 0.05, -0.05],
            "est_standard_devs": [0.005, 0.005, 0.005, 0.005],
        },
    }
    d = plot_scaling_models(PhysicalScalingModel.from_dict(physical_dict))
    assert "smooth_scale_model" in d
    assert "absorption_surface" in d
    assert "absorption_parameters" in d
    assert d["smooth_scale_model"]["data"][0] != []
    assert d["absorption_parameters"]["data"][0] != []


def test_normal_probability_plot():
    data = {"delta_hl": np.arange(20, dtype=float)}
    d = normal_probability_plot(data)
    assert "normal_distribution_plot" in d


def test_plot_outliers():
    """Test outlier plot, for standard and null data."""
    data = {"x": [1.0, 2.0], "y": [0.0, 1.0], "z": [1.0, 1.0], "image_size": [100, 200]}
    d = plot_outliers(data)
    assert "outliers_vs_z" in d
    assert "outlier_xy_positions" in d

    data = {"x": [], "y": [], "z": [], "image_size": [100, 200]}
    d = plot_outliers(data)
    assert "outliers_vs_z" in d
    assert "outlier_xy_positions" in d



from __future__ import annotations

from pathlib import Path

import procrunner
import pytest

from dxtbx.model.experiment_list import ExperimentListFactory
from libtbx import phil

from dials.algorithms.refinement import RefinerFactory
from dials.algorithms.refinement.engine import Journal
from dials.algorithms.refinement.refiner import phil_scope as refiner_phil_scope
from dials.array_family import flex


def test_joint_refinement(dials_regression, tmp_path):
    """A basic test of joint refinement of the CS-PAD detector at hierarchy level 2
    with 300 crystals."""

    bevington = pytest.importorskip("scitbx.examples.bevington")
    if not hasattr(bevington, "non_linear_ls_eigen_wrapper"):
        pytest.skip("Skipping test as SparseLevMar engine not available")

    data_dir = Path(dials_regression) / "refinement_test_data" / "xfel_metrology"

    # Do refinement and load the history
    result = procrunner.run(
        [
            "dials.refine",
            data_dir / "benchmark_level2d.json",
            data_dir / "benchmark_level2d.pickle",
            data_dir / "refine.phil",
            "history=history.json",
        ],
        working_directory=tmp_path,
    )
    assert not result.returncode and not result.stderr

    # there are plenty of things we could do with the refinement history, but
    # here just check that final RMSDs are low enough
    history = Journal.from_json_file(tmp_path / "history.json")
    final_rmsd = history["rmsd"][-1]
    assert final_rmsd[0] < 0.0354
    assert final_rmsd[1] < 0.0406
    assert final_rmsd[2] < 0.0018

    # also check that the used_in_refinement flag got set correctly
    rt = flex.reflection_table.from_file(tmp_path / "refined.refl")
    uir = rt.get_flags(rt.flags.used_in_refinement)
    assert uir.count(True) == history["num_reflections"][-1]


def test_constrained_refinement(dials_regression, tmp_path):
    """Do constrained refinement, checking that a panel group with no data
    on it still moves with its partners in the constraint.
    See https://github.com/dials/dials/issues/990"""

    bevington = pytest.importorskip("scitbx.examples.bevington")
    if not hasattr(bevington, "non_linear_ls_eigen_wrapper"):
        pytest.skip("Skipping test as SparseLevMar engine not available")

    data_dir = Path(dials_regression) / "refinement_test_data" / "xfel_metrology"

    # Load experiments and reflections
    refl = flex.reflection_table.from_file(data_dir / "benchmark_level2d.pickle")
    expt = ExperimentListFactory.from_json_file(data_dir / "benchmark_level2d.json")

    # There are zero reflections on some panels, so these will only move via constraints
    for i in [8, 10, 11, 26, 27, 40, 42, 43, 56, 58, 59]:
        assert (refl["panel"] == i).count(True) == 0

    # Get parameters, combining refine.phil with constraints that enforce distances to move in lockstep
    refine_phil = phil.parse(data_dir.joinpath("refine.phil").read_text())
    constraint_phil = phil.parse(
        """
refinement {
  parameterisation {
    detector {
      fix_list=Tau2,Tau3
      constraints {
        parameter=Dist
      }
    }
  }
}
"""
    )
    params = refiner_phil_scope.fetch(sources=[refine_phil, constraint_phil]).extract()

    detector = expt.detectors()[0]
    initial_distances = [p.get_distance() for p in detector]

    # Set up a refiner
    refiner = RefinerFactory.from_parameters_data_experiments(params, refl, expt)

    refiner.run()
    detector = refiner.get_experiments().detectors()[0]

    final_distances = [p.get_distance() for p in detector]

    # The shifts between initial and final distances should all be equal
    dist_diff = [a - b for a, b in zip(final_distances, initial_distances)]
    for d in dist_diff[1:]:
        assert d == pytest.approx(dist_diff[0])



from __future__ import annotations

import math


def test_manhattan():
    from scitbx.array_family import flex

    from dials.algorithms.image.filter import manhattan_distance

    data = flex.bool(flex.grid(100, 100), True)

    for j in range(100):
        for i in range(100):
            if math.sqrt((j - 50) ** 2 + (i - 50) ** 2) <= 10.5:
                data[j, i] = False

    distance = manhattan_distance(data, True)

    M = distance[1:-1, 1:-1]
    D = data[1:-1, 1:-1]

    selection = D.as_1d().select(M.as_1d() == 0)
    assert selection.all_eq(True)

    while True:
        N = data[0:-2, 1:-1]
        S = data[2:, 1:-1]
        E = data[1:-1, 0:-2]
        W = data[1:-1, 2:]
        selection = M.as_1d() == 1
        neighbours = (
            N.select(selection)
            | S.select(selection)
            | E.select(selection)
            | W.select(selection)
        )
        assert neighbours.all_eq(True)
        indices = flex.size_t(range(len(D))).select(selection)
        for i in indices:
            D[i] = True
        data[1:-1, 1:-1] = D
        M -= 1
        if (M > 0).count(True) == 0:
            break


def test_chebyshev():
    from scitbx.array_family import flex

    from dials.algorithms.image.filter import chebyshev_distance

    data = flex.bool(flex.grid(100, 100), True)

    for j in range(100):
        for i in range(100):
            if math.sqrt((j - 50) ** 2 + (i - 50) ** 2) <= 10.5:
                data[j, i] = False

    distance = chebyshev_distance(data, True)

    M = distance[1:-1, 1:-1]
    D = data[1:-1, 1:-1]
    selection = D.as_1d().select(M.as_1d() == 0)
    assert selection.all_eq(True)

    while True:
        N = data[0:-2, 1:-1]
        S = data[2:, 1:-1]
        E = data[1:-1, 0:-2]
        W = data[1:-1, 2:]
        NE = data[0:-2, 0:-2]
        NW = data[0:-2, 2:]
        SE = data[2:, 0:-2]
        SW = data[2:, 2:]
        selection = M.as_1d() == 1
        neighbours = (
            N.select(selection)
            | S.select(selection)
            | E.select(selection)
            | W.select(selection)
            | NE.select(selection)
            | NW.select(selection)
            | SE.select(selection)
            | SW.select(selection)
        )
        assert neighbours.all_eq(True)
        indices = flex.size_t(range(len(D))).select(selection)
        for i in indices:
            D[i] = True
        data[1:-1, 1:-1] = D
        M -= 1
        if (M > 0).count(True) == 0:
            break


def test_chebyshev_small():
    from scitbx.array_family import flex

    from dials.algorithms.image.filter import chebyshev_distance

    data = flex.bool(flex.grid(5, 5), True)
    data[1, 1] = False

    distance = chebyshev_distance(data, False)

    known = flex.int(
        [1, 1, 1, 2, 3, 1, 0, 1, 2, 3, 1, 1, 1, 2, 3, 2, 2, 2, 2, 3, 3, 3, 3, 3, 3]
    )
    known.reshape(distance.accessor())

    assert (known == distance).count(False) == 0



from __future__ import annotations

import os

import pytest

from cctbx import uctbx

from .test_index import run_indexing


@pytest.mark.xfail
def test_run(dials_regression, tmpdir):
    expected_unit_cell = uctbx.unit_cell(
        (11.624, 13.550, 30.103, 89.964, 93.721, 90.132)
    )
    expected_rmsds = (0.039, 0.035, 0.002)

    experiments_old = os.path.join(
        dials_regression, "indexing_test_data", "phi_scan", "datablock_old.json"
    )
    experiments_new = os.path.join(
        dials_regression, "indexing_test_data", "phi_scan", "datablock.json"
    )
    strong_pickle = os.path.join(
        dials_regression, "indexing_test_data", "phi_scan", "strong.pickle"
    )

    from dxtbx.serialize import load

    imageset_old = load.experiment_list(
        experiments_old, check_format=False
    ).imagesets()[0]
    imageset_new = load.experiment_list(
        experiments_new, check_format=False
    ).imagesets()[0]

    gonio_old = imageset_old.get_goniometer()
    gonio_new = imageset_new.get_goniometer()

    assert gonio_old.get_rotation_axis() == pytest.approx(
        (0.7497646259807715, -0.5517923303436749, 0.36520984351713554)
    )
    assert gonio_old.get_setting_rotation() == pytest.approx(
        (1.0, 0.0, 0.0, 0.0, 1.0, 0.0, 0.0, 0.0, 1.0)
    )
    assert gonio_old.get_fixed_rotation() == pytest.approx(
        (
            0.7497646259807748,
            -0.20997265900532208,
            -0.6275065641872948,
            -0.5517923303436731,
            0.3250014637526764,
            -0.7680490041218182,
            0.3652098435171313,
            0.9221092836691605,
            0.12781329809272568,
        )
    )

    assert gonio_new.get_rotation_axis() == pytest.approx(gonio_old.get_rotation_axis())
    assert gonio_new.get_rotation_axis_datum() == pytest.approx((1, 0, 0))
    assert gonio_new.get_setting_rotation() == pytest.approx(
        (
            0.7497646259807705,
            -0.20997265900532142,
            -0.6275065641873,
            -0.5517923303436786,
            0.3250014637526763,
            -0.768049004121814,
            0.3652098435171315,
            0.9221092836691607,
            0.12781329809272335,
        )
    )
    assert gonio_new.get_fixed_rotation() == pytest.approx(
        (1.0, 0.0, 0.0, 0.0, 1.0, 0.0, 0.0, 0.0, 1.0)
    )

    result_old = run_indexing(
        strong_pickle,
        experiments_old,
        tmpdir,
        extra_args=[],
        expected_unit_cell=expected_unit_cell,
        expected_rmsds=expected_rmsds,
        expected_hall_symbol=" P 1",
    )

    result_new = run_indexing(
        strong_pickle,
        experiments_new,
        tmpdir,
        extra_args=[],
        expected_unit_cell=expected_unit_cell,
        expected_rmsds=expected_rmsds,
        expected_hall_symbol=" P 1",
    )

    assert result_old.rmsds == pytest.approx(result_new.rmsds, abs=1e-6)
    assert result_old.experiments[
        0
    ].crystal.get_unit_cell().parameters() == pytest.approx(
        result_new.experiments[0].crystal.get_unit_cell().parameters(), abs=1e-6
    )

    # Now test refinement gradients are correct
    from dxtbx.model.experiment_list import Experiment, ExperimentList

    old_exps = ExperimentList(
        [
            Experiment(
                beam=imageset_old.get_beam(),
                detector=imageset_old.get_detector(),
                goniometer=gonio_old,
                scan=imageset_old.get_scan(),
                crystal=result_old.experiments[0].crystal,
                imageset=None,
            )
        ]
    )
    new_exps = ExperimentList(
        [
            Experiment(
                beam=imageset_new.get_beam(),
                detector=imageset_new.get_detector(),
                goniometer=gonio_new,
                scan=imageset_new.get_scan(),
                crystal=result_new.experiments[0].crystal,
                imageset=None,
            )
        ]
    )

    from libtbx.phil import parse

    from dials.algorithms.refinement.refiner import phil_scope

    params = phil_scope.fetch(source=parse("")).extract()
    from dials.algorithms.refinement.refiner import RefinerFactory

    refiner_old = RefinerFactory.from_parameters_data_experiments(
        params, result_old.indexed_reflections, old_exps
    )
    refiner_new = RefinerFactory.from_parameters_data_experiments(
        params, result_new.indexed_reflections, new_exps
    )

    # Analytical gradients should be approximately the same in either case
    an_grads_old = refiner_old._pred_param.get_gradients(refiner_old.get_matches())
    an_grads_new = refiner_new._pred_param.get_gradients(refiner_new.get_matches())
    for g1, g2 in zip(an_grads_old, an_grads_new):
        assert g1["dX_dp"] == pytest.approx(g2["dX_dp"], abs=1.0e-6)
        assert g1["dY_dp"] == pytest.approx(g2["dY_dp"], abs=1.0e-6)
        assert g1["dphi_dp"] == pytest.approx(g2["dphi_dp"], abs=1.0e-6)

    # Analytical gradients should be approximately equal to finite difference
    # gradients in either case
    fd_grads_old = calc_fd_grads(refiner_old)
    for g1, g2 in zip(fd_grads_old, an_grads_old):
        assert g1["dX_dp"] == pytest.approx(g2["dX_dp"], abs=5.0e-6)
        assert g1["dY_dp"] == pytest.approx(g2["dY_dp"], abs=5.0e-6)
        assert g1["dphi_dp"] == pytest.approx(g2["dphi_dp"], abs=5.0e-6)
    fd_grads_new = calc_fd_grads(refiner_new)
    for g1, g2 in zip(fd_grads_new, an_grads_new):
        assert g1["dX_dp"] == pytest.approx(g2["dX_dp"], abs=5.0e-6)
        assert g1["dY_dp"] == pytest.approx(g2["dY_dp"], abs=5.0e-6)
        assert g1["dphi_dp"] == pytest.approx(g2["dphi_dp"], abs=5.0e-6)


def calc_fd_grads(refiner):

    p_vals = refiner._pred_param.get_param_vals()
    deltas = [1.0e-7] * len(p_vals)

    fd_grads = []
    for i, val in enumerate(p_vals):
        p_vals[i] -= deltas[i] / 2.0
        refiner._pred_param.set_param_vals(p_vals)

        refiner._target.predict()

        rev_state = refiner.get_matches()["xyzcal.mm"].deep_copy()

        p_vals[i] += deltas[i]
        refiner._pred_param.set_param_vals(p_vals)

        refiner._target.predict()

        fwd_state = refiner.get_matches()["xyzcal.mm"].deep_copy()
        p_vals[i] = val

        fd = fwd_state - rev_state
        x_grads, y_grads, phi_grads = fd.parts()
        x_grads /= deltas[i]
        y_grads /= deltas[i]
        phi_grads /= deltas[i]

        fd_grads.append({"dX_dp": x_grads, "dY_dp": y_grads, "dphi_dp": phi_grads})

    # return to the initial state
    refiner._pred_param.set_param_vals(p_vals)

    return fd_grads



from __future__ import annotations

from math import exp
from random import randint

import pytest
from numpy.random import poisson

from scitbx.array_family import flex

from dials.algorithms.image.threshold import (
    DispersionExtendedThreshold,
    DispersionExtendedThresholdDebug,
    DispersionThreshold,
    DispersionThresholdDebug,
)


class Test:
    def setup_class(self):
        from scitbx.array_family import flex

        spot = flex.double(flex.grid(11, 11))
        for j in range(11):
            for i in range(11):
                spot[j, i] = exp(-((j - 5) ** 2 + (i - 5) ** 2) / 2**2)

        self.image = flex.double(flex.grid(2000, 2000))
        for n in range(200):
            x = randint(0, 2000)
            y = randint(0, 2000)
            for j in range(0, 11):
                for i in range(0, 11):
                    xx = x + i - 5
                    yy = y + j - 5
                    if xx >= 0 and yy >= 0 and xx < 2000 and yy < 2000:
                        self.image[yy, xx] = poisson(100 * spot[j, i])

        background = flex.double(list(poisson(5, 2000 * 2000)))
        background.reshape(flex.grid(2000, 2000))

        self.image += background

        # Create an image
        self.mask = flex.random_bool(2000 * 2000, 0.99)
        self.mask.reshape(flex.grid(2000, 2000))
        self.gain = flex.random_double(2000 * 2000) + 1.0
        self.gain.reshape(flex.grid(2000, 2000))
        self.size = (3, 3)
        self.min_count = 2

    def test_niblack(self):
        from dials.algorithms.image.threshold import niblack

        n_sigma = 3
        niblack(self.image, self.size, n_sigma)

    def test_sauvola(self):
        from dials.algorithms.image.threshold import sauvola

        k = 3
        r = 3
        sauvola(self.image, self.size, k, r)

    def test_index_of_dispersion(self):
        from dials.algorithms.image.threshold import index_of_dispersion

        n_sigma = 3
        index_of_dispersion(self.image, self.size, n_sigma)

    def test_index_of_dispersion_masked(self):
        from dials.algorithms.image.threshold import index_of_dispersion_masked

        n_sigma = 3
        index_of_dispersion_masked(
            self.image, self.mask, self.size, self.min_count, n_sigma
        )

    def test_gain(self):
        from dials.algorithms.image.threshold import gain

        n_sigma = 3
        gain(self.image, self.mask, self.gain, self.size, self.min_count, n_sigma)

    def test_dispersion(self):
        from dials.algorithms.image.threshold import dispersion

        nsig_b = 3
        nsig_s = 3
        dispersion(self.image, self.mask, self.size, nsig_b, nsig_s, self.min_count)

    def test_dispersion_w_gain(self):
        from dials.algorithms.image.threshold import dispersion_w_gain

        nsig_b = 3
        nsig_s = 3
        result1 = dispersion_w_gain(
            self.image, self.mask, self.gain, self.size, nsig_b, nsig_s, self.min_count
        )

        # scaling both the image and the gain should not affect the result
        result2 = dispersion_w_gain(
            2.0 * self.image,
            self.mask,
            2.0 * self.gain,
            self.size,
            nsig_b,
            nsig_s,
            self.min_count,
        )
        assert result1 == result2

        # should get the same result as dispersion if the gain is unity
        from dials.algorithms.image.threshold import dispersion

        result3 = dispersion(
            self.image, self.mask, self.size, nsig_b, nsig_s, self.min_count
        )
        result4 = dispersion_w_gain(
            self.image,
            self.mask,
            (0 * self.gain + 1),
            self.size,
            nsig_b,
            nsig_s,
            self.min_count,
        )
        assert result3 == result4

    def test_dispersion_debug(self):
        from dials.algorithms.image.threshold import (
            DispersionThresholdDebug,
            dispersion,
            dispersion_w_gain,
        )

        nsig_b = 3
        nsig_s = 3
        result1 = dispersion(
            self.image, self.mask, self.size, nsig_b, nsig_s, self.min_count
        )
        debug = DispersionThresholdDebug(
            self.image, self.mask, self.size, nsig_b, nsig_s, 0, self.min_count
        )
        result2 = debug.final_mask()
        assert result1.all_eq(result2)

        result3 = dispersion_w_gain(
            self.image, self.mask, self.gain, self.size, nsig_b, nsig_s, self.min_count
        )
        debug = DispersionThresholdDebug(
            self.image,
            self.mask,
            self.gain,
            self.size,
            nsig_b,
            nsig_s,
            0,
            self.min_count,
        )
        result4 = debug.final_mask()
        assert result3 == result4

    def test_dispersion_threshold(self):
        from dials.algorithms.image.threshold import (
            DispersionThreshold,
            dispersion,
            dispersion_w_gain,
        )
        from dials.array_family import flex

        nsig_b = 3
        nsig_s = 3
        algorithm = DispersionThreshold(
            self.image.all(), self.size, nsig_b, nsig_s, 0, self.min_count
        )

        result1 = dispersion(
            self.image, self.mask, self.size, nsig_b, nsig_s, self.min_count
        )

        result2 = dispersion_w_gain(
            self.image, self.mask, self.gain, self.size, nsig_b, nsig_s, self.min_count
        )

        result3 = flex.bool(flex.grid(self.image.all()))
        result4 = flex.bool(flex.grid(self.image.all()))
        algorithm(self.image, self.mask, result3)
        algorithm(self.image, self.mask, self.gain, result4)
        assert result1 == result3
        assert result2 == result4

    def test_dispersion_extended_threshold(self):
        from dials.algorithms.image.threshold import (
            DispersionExtendedThreshold,
            DispersionExtendedThresholdDebug,
        )
        from dials.array_family import flex

        nsig_b = 3
        nsig_s = 3
        algorithm = DispersionExtendedThreshold(
            self.image.all(), self.size, nsig_b, nsig_s, 0, self.min_count
        )
        result1 = flex.bool(flex.grid(self.image.all()))
        result2 = flex.bool(flex.grid(self.image.all()))
        algorithm(self.image, self.mask, result1)
        algorithm(self.image, self.mask, self.gain, result2)

        debug = DispersionExtendedThresholdDebug(
            self.image, self.mask, self.size, nsig_b, nsig_s, 0, self.min_count
        )
        result3 = debug.final_mask()

        assert result1.all_eq(result3)

        debug = DispersionExtendedThresholdDebug(
            self.image,
            self.mask,
            self.gain,
            self.size,
            nsig_b,
            nsig_s,
            0,
            self.min_count,
        )
        result4 = debug.final_mask()
        assert result2 == result4

    @pytest.mark.parametrize(
        "algorithm", [DispersionThreshold, DispersionExtendedThreshold]
    )
    def test_dispersion_algorithm_symmetry(self, algorithm):

        nsig_b = 3
        nsig_s = 3

        thresholder = algorithm(
            self.image.all(), (3, 3), nsig_b, nsig_s, 0, self.min_count
        )
        result1 = flex.bool(flex.grid(self.image.all()))
        thresholder(self.image, self.mask, result1)

        # Ugh. Is there an easier way to transpose a 2d flex.bool?
        def transpose_a_flex_bool(arr):
            arr_t = arr.as_1d().as_int()
            arr_t.reshape(flex.grid(arr.all()))
            arr_t.matrix_transpose_in_place()
            new_dim = arr_t.all()
            arr_t = arr_t.as_1d().as_bool()
            arr_t.reshape(flex.grid(new_dim))
            return arr_t

        image_t = self.image.matrix_transpose()
        mask_t = transpose_a_flex_bool(self.mask)
        thresholder = algorithm(
            image_t.all(), (3, 3), nsig_b, nsig_s, 0, self.min_count
        )
        result2 = flex.bool(flex.grid(image_t.all()))
        thresholder(image_t, mask_t, result2)
        result2_t = transpose_a_flex_bool(result2)

        assert (result1 == result2_t).all_eq(True)

    @pytest.mark.parametrize(
        "algorithm", [DispersionThresholdDebug, DispersionExtendedThresholdDebug]
    )
    def test_dispersion_debug_algorithm_symmetry(self, algorithm):

        nsig_b = 3
        nsig_s = 3

        threshold = algorithm(
            self.image, self.mask, (3, 3), nsig_b, nsig_s, 0, self.min_count
        )
        result1 = threshold.final_mask()

        # Ugh. Is there an easier way to transpose a 2d flex.bool?
        def transpose_a_flex_bool(arr):
            arr_t = arr.as_1d().as_int()
            arr_t.reshape(flex.grid(arr.all()))
            arr_t.matrix_transpose_in_place()
            new_dim = arr_t.all()
            arr_t = arr_t.as_1d().as_bool()
            arr_t.reshape(flex.grid(new_dim))
            return arr_t

        image_t = self.image.matrix_transpose()
        mask_t = transpose_a_flex_bool(self.mask)
        threshold = algorithm(
            image_t, mask_t, (3, 3), nsig_b, nsig_s, 0, self.min_count
        )
        result2 = threshold.final_mask()
        result2_t = transpose_a_flex_bool(result2)

        assert (result1 == result2_t).all_eq(True)



from __future__ import annotations

import copy
import logging
import os

import dxtbx.model
import libtbx.phil
from cctbx.miller import map_to_asu
from rstbx.cftbx.coordinate_frame_helpers import align_reference_frame
from scitbx import matrix

from dials.array_family import flex
from dials.util import Sorry
from dials.util.filter_reflections import (
    FilteringReductionMethods,
    filter_reflection_table,
)

try:
    from typing import Tuple
except ImportError:
    pass

logger = logging.getLogger(__name__)


def export_xds_ascii(integrated_data, experiment_list, params, var_model=(1, 0)):
    """Export data from integrated_data corresponding to experiment_list to
    an XDS_ASCII.HKL formatted text file."""

    if len(experiment_list) == 1:
        experiment_data = integrated_data.select(integrated_data["id"] >= 0)
        _export_experiment(
            params.xds_ascii.hklout,
            experiment_data,
            experiment_list[0],
            params,
            var_model,
        )
    else:
        for i, experiment in enumerate(experiment_list):
            experiment_data = integrated_data.select(integrated_data["id"] == i)
            name, ext = os.path.splitext(params.xds_ascii.hklout)
            filename = name + f"_{i}" + ext
            _export_experiment(filename, experiment_data, experiment, params, var_model)


def _export_experiment(filename, integrated_data, experiment, params, var_model=(1, 0)):
    # type: (str, flex.reflection_table, dxtbx.model.Experiment, libtbx.phil.scope_extract, Tuple)
    """Export a single experiment to an XDS_ASCII.HKL format file.

    Args:
        filename: The file to write to
        integrated_data: The reflection table, pre-selected to one experiment
        experiment: The experiment list entry to export
        params: The PHIL configuration object
        var_model:
    """
    # export for xds_ascii should only be for non-scaled reflections
    assert any(
        i in integrated_data for i in ["intensity.sum.value", "intensity.prf.value"]
    )
    # Handle requesting profile intensities (default via auto) but no column
    if "profile" in params.intensity and "intensity.prf.value" not in integrated_data:
        raise Sorry(
            "Requested profile intensity data but only summed present. Use intensity=sum."
        )

    integrated_data = filter_reflection_table(
        integrated_data,
        intensity_choice=params.intensity,
        partiality_threshold=params.mtz.partiality_threshold,
        combine_partials=params.mtz.combine_partials,
        min_isigi=params.mtz.min_isigi,
        filter_ice_rings=params.mtz.filter_ice_rings,
        d_min=params.mtz.d_min,
    )

    # calculate the scl = lp/dqe correction for outputting but don't apply it as
    # it has already been applied in filter_reflection_table
    (
        integrated_data,
        scl,
    ) = FilteringReductionMethods.calculate_lp_qe_correction_and_filter(integrated_data)

    # sort data before output
    nref = len(integrated_data["miller_index"])
    indices = flex.size_t_range(nref)

    unique = copy.deepcopy(integrated_data["miller_index"])

    map_to_asu(experiment.crystal.get_space_group().type(), False, unique)

    perm = sorted(indices, key=lambda k: unique[k])
    integrated_data = integrated_data.select(flex.size_t(perm))

    if experiment.goniometer is None:
        print("Warning: No goniometer. Experimentally exporting with (1 0 0) axis")

    unit_cell = experiment.crystal.get_unit_cell()

    if experiment.scan is None:
        print("Warning: No Scan. Experimentally exporting no-oscillation values")
        image_range = (1, 1)
        phi_start, phi_range = 0.0, 0.0
    else:
        image_range = experiment.scan.get_image_range()
        phi_start, phi_range = experiment.scan.get_image_oscillation(image_range[0])

    # gather the required information for the reflection file

    nref = len(integrated_data["miller_index"])

    miller_index = integrated_data["miller_index"]

    # profile correlation
    if "profile.correlation" in integrated_data:
        prof_corr = 100.0 * integrated_data["profile.correlation"]
    else:
        prof_corr = flex.double(nref, 100.0)

    # partiality
    if "partiality" in integrated_data:
        partiality = 100 * integrated_data["partiality"]
    else:
        prof_corr = flex.double(nref, 100.0)

    if "intensity.sum.value" in integrated_data:
        I = integrated_data["intensity.sum.value"]
        V = integrated_data["intensity.sum.variance"]
        assert V.all_gt(0)
        V = var_model[0] * (V + var_model[1] * I * I)
        sigI = flex.sqrt(V)
    else:
        I = integrated_data["intensity.prf.value"]
        V = integrated_data["intensity.prf.variance"]
        assert V.all_gt(0)
        V = var_model[0] * (V + var_model[1] * I * I)
        sigI = flex.sqrt(V)

    fout = open(filename, "w")

    # first write the header - in the "standard" coordinate frame...

    panel = experiment.detector[0]
    fast = panel.get_fast_axis()
    slow = panel.get_slow_axis()
    Rd = align_reference_frame(fast, (1, 0, 0), slow, (0, 1, 0))
    print("Coordinate change:")
    print("%5.2f %5.2f %5.2f\n%5.2f %5.2f %5.2f\n%5.2f %5.2f %5.2f\n" % Rd.elems)

    fast = Rd * fast
    slow = Rd * slow

    qx, qy = panel.get_pixel_size()
    nx, ny = panel.get_image_size()
    distance = matrix.col(Rd * panel.get_origin()).dot(
        matrix.col(Rd * panel.get_normal())
    )
    org = Rd * (
        matrix.col(panel.get_origin()) - distance * matrix.col(panel.get_normal())
    )
    orgx = -org.dot(fast) / qx
    orgy = -org.dot(slow) / qy

    UB = Rd * matrix.sqr(experiment.crystal.get_A())
    real_space_ABC = UB.inverse().elems

    if experiment.goniometer is not None:
        axis = Rd * experiment.goniometer.get_rotation_axis()
    else:
        axis = Rd * (1, 0, 0)

    beam = Rd * experiment.beam.get_s0()
    cell_fmt = "%9.3f %9.3f %9.3f %7.3f %7.3f %7.3f"
    axis_fmt = "%9.3f %9.3f %9.3f"

    fout.write(
        "\n".join(
            [
                "!FORMAT=XDS_ASCII    MERGE=FALSE    FRIEDEL'S_LAW=TRUE",
                "!Generated by dials.export",
                "!DATA_RANGE= %d %d" % image_range,
                "!ROTATION_AXIS= %9.6f %9.6f %9.6f" % axis.elems,
                "!OSCILLATION_RANGE= %f" % phi_range,
                "!STARTING_ANGLE= %f" % phi_start,
                "!STARTING_FRAME= %d" % image_range[0],
                "!SPACE_GROUP_NUMBER= %d"
                % experiment.crystal.get_space_group().type().number(),
                f"!UNIT_CELL_CONSTANTS= {cell_fmt % unit_cell.parameters()}",
                f"!UNIT_CELL_A-AXIS= {axis_fmt % real_space_ABC[0:3]}",
                f"!UNIT_CELL_B-AXIS= {axis_fmt % real_space_ABC[3:6]}",
                f"!UNIT_CELL_C-AXIS= {axis_fmt % real_space_ABC[6:9]}",
                f"!X-RAY_WAVELENGTH= {experiment.beam.get_wavelength():f}",
                "!INCIDENT_BEAM_DIRECTION= %f %f %f" % beam.elems,
                "!NX= %d NY= %d QX= %f QY= %f" % (nx, ny, qx, qy),
                f"!ORGX= {orgx:9.2f} ORGY= {orgy:9.2f}",
                f"!DETECTOR_DISTANCE= {distance:8.3f}",
                "!DIRECTION_OF_DETECTOR_X-AXIS= %9.5f %9.5f %9.5f" % fast.elems,
                "!DIRECTION_OF_DETECTOR_Y-AXIS= %9.5f %9.5f %9.5f" % slow.elems,
                "!VARIANCE_MODEL= %7.3e %7.3e" % var_model,
                "!NUMBER_OF_ITEMS_IN_EACH_DATA_RECORD=12",
                "!ITEM_H=1",
                "!ITEM_K=2",
                "!ITEM_L=3",
                "!ITEM_IOBS=4",
                "!ITEM_SIGMA(IOBS)=5",
                "!ITEM_XD=6",
                "!ITEM_YD=7",
                "!ITEM_ZD=8",
                "!ITEM_RLP=9",
                "!ITEM_PEAK=10",
                "!ITEM_CORR=11",
                "!ITEM_PSI=12",
                "!END_OF_HEADER",
                "",
            ]
        )
    )

    # then write the data records

    s0 = Rd * matrix.col(experiment.beam.get_s0())

    for j in range(nref):
        x, y, z = integrated_data["xyzcal.px"][j]
        phi = phi_start + z * phi_range
        h, k, l = miller_index[j]
        X = (UB * (h, k, l)).rotate(axis, phi, deg=True)
        s = s0 + X
        g = s.cross(s0).normalize()

        # find component of beam perpendicular to f, e
        e = -(s + s0).normalize()
        if h == k and k == l:
            u = (h, -h, 0)
        else:
            u = (k - l, l - h, h - k)
        q = (
            (matrix.col(u).transpose() * UB.inverse())
            .normalize()
            .transpose()
            .rotate(axis, phi, deg=True)
        )

        psi = q.angle(g, deg=True)
        if q.dot(e) < 0:
            psi *= -1

        fout.write(
            "%d %d %d %f %f %f %f %f %f %.1f %.1f %f\n"
            % (
                h,
                k,
                l,
                I[j],
                sigI[j],
                x,
                y,
                z,
                scl[j],
                partiality[j],
                prof_corr[j],
                psi,
            )
        )

    fout.write("!END_OF_DATA\n")
    fout.close()
    logger.info("Output %d reflections to %s", nref, filename)
